"""
Use of this source code is governed by the MIT license found in the LICENSE file.

General node object to control associated plugwise nodes like: Circle+, Circle, Scan, Stealth
"""
from datetime import datetime
from plugwise.constants import (
    HA_SWITCH,
    HW_MODELS,
    SENSOR_AVAILABLE,
    SENSOR_RSSI_IN,
    SENSOR_RSSI_OUT,
    SENSOR_PING,
    SWITCH_RELAY,
    UTF8_DECODE,
)
from plugwise.message import PlugwiseMessage
from plugwise.messages.responses import (
    NodeFeaturesResponse,
    NodeInfoResponse,
    NodePingResponse,
)
from plugwise.messages.requests import (
    NodeFeaturesRequest,
    NodeInfoRequest,
    NodePingRequest,
)
from plugwise.util import validate_mac


class PlugwiseNode(object):
    """ Base class for a Plugwise node """

    def __init__(self, mac, address, stick):
        mac = mac.upper()
        if validate_mac(mac) == False:
            self.stick.logger.debug(
                "MAC address is in unexpected format: %s",
                str(mac),
            )
        self.mac = bytes(mac, encoding=UTF8_DECODE)
        self.stick = stick
        self.categories = ()
        self.sensors = ()
        self.switches = ()
        self._address = address
        self._callbacks = {}
        self.last_update = None
        self.last_request = None
        self._available = False
        self.in_RSSI = None
        self.out_RSSI = None
        self.ping_ms = None
        self._node_type = None
        self._hardware_version = None
        self._firmware_version = None
        self._relay_state = False
        self._last_log_address = None
        self._last_log_collected = False
        self._last_info_message = None
        self._features = None

    def get_node_type(self) -> str:
        """Return hardware model"""
        if self._hardware_version:
            hw_model = HW_MODELS.get(self._hardware_version[4:10], None)
            if hw_model:
                return hw_model
            else:
                # Try again with reversed order
                hw_model = HW_MODELS.get(
                    self._hardware_version[-2:]
                    + self._hardware_version[-4:-2]
                    + self._hardware_version[-6:-4],
                    None,
                )
                if hw_model:
                    return hw_model
        return "Unknown"

    def is_sed(self) -> bool:
        """ Return True if node SED (battery powered)"""
        return False

    def get_categories(self) -> tuple:
        """ Return Home Assistant catagories supported by plugwise node """
        return self.categories

    def get_sensors(self) -> tuple:
        """ Return sensors supported by plugwise node """
        return self.sensors

    def get_switches(self) -> tuple:
        """ Return switches supported by plugwise node """
        return self.switches

    def get_available(self) -> bool:
        """ Return current network state of plugwise node """
        return self._available

    def set_available(self, state, request_info=False):
        """ Set current network state of plugwise node """
        if state == True:
            if self._available == False:
                self._available = True
                self.stick.logger.debug(
                    "Mark node %s available",
                    self.get_mac(),
                )
                self.do_callback(SENSOR_AVAILABLE["id"])
                if request_info:
                    self._request_info()
        else:
            if self._available == True:
                self._available = False
                self.stick.logger.debug(
                    "Mark node %s unavailable",
                    self.get_mac(),
                )
                self.do_callback(SENSOR_AVAILABLE["id"])

    def get_mac(self) -> str:
        """Return mac address"""
        return self.mac.decode(UTF8_DECODE)

    def get_name(self) -> str:
        """Return unique name"""
        return self.get_node_type() + " (" + str(self._address) + ")"

    def get_hardware_version(self) -> str:
        """Return hardware version"""
        if self._hardware_version != None:
            return self._hardware_version
        return "Unknown"

    def get_firmware_version(self) -> str:
        """Return firmware version"""
        if self._firmware_version != None:
            return str(self._firmware_version)
        return "Unknown"

    def get_last_update(self) -> datetime:
        """Return  version"""
        return self.last_update

    def get_in_RSSI(self) -> int:
        """Return inbound RSSI level"""
        if self.in_RSSI != None:
            return self.in_RSSI
        return 0

    def get_out_RSSI(self) -> int:
        """Return outbound RSSI level"""
        if self.out_RSSI != None:
            return self.out_RSSI
        return 0

    def get_ping(self) -> int:
        """Return ping roundtrip"""
        if self.ping_ms != None:
            return self.ping_ms
        return 0

    def _request_info(self, callback=None):
        """ Request info from node"""
        self.stick.send(
            NodeInfoRequest(self.mac),
            callback,
        )

    def _request_features(self, callback=None):
        """ Request supported features for this node"""
        self.stick.send(
            NodeFeaturesRequest(self.mac),
            callback,
        )

    def ping(self, callback=None):
        """ Ping node"""
        self.stick.send(
            NodePingRequest(self.mac),
            callback,
        )

    def on_message(self, message):
        """
        Process received message
        """
        assert isinstance(message, PlugwiseMessage)
        if message.mac == self.mac:
            if message.timestamp != None:
                self.stick.logger.debug(
                    "Last update %s of node %s, last message %s",
                    str(self.last_update),
                    self.get_mac(),
                    str(message.timestamp),
                )
                self.last_update = message.timestamp
            if isinstance(message, NodePingResponse):
                self._process_ping_response(message)
            elif isinstance(message, NodeInfoResponse):
                self._process_info_response(message)
            elif isinstance(message, NodeFeaturesResponse):
                self._process_features_response(message)
            else:
                self._on_message(message)
                self.set_available(True)
        else:
            self.stick.logger.debug(
                "Skip message, mac of node (%s) != mac at message (%s)",
                message.mac.decode(UTF8_DECODE),
                self.get_mac(),
            )

    def _on_message(self, message):
        pass

    def subscribe_callback(self, callback, sensor) -> bool:
        """ Subscribe callback to execute when state change happens """
        if sensor in self.sensors:
            if sensor not in self._callbacks:
                self._callbacks[sensor] = []
            self._callbacks[sensor].append(callback)
            return True
        return False

    def unsubscribe_callback(self, callback, sensor):
        """ Unsubscribe callback to execute when state change happens """
        if sensor in self._callbacks:
            self._callbacks[sensor].remove(callback)

    def do_callback(self, sensor):
        """ Execute callbacks registered for specified callback type """
        if sensor in self._callbacks:
            for callback in self._callbacks[sensor]:
                try:
                    callback(None)
                except Exception as e:
                    self.stick.logger.error(
                        "Error while executing all callback : %s",
                        e,
                    )

    def _process_ping_response(self, message):
        """ Process ping response message"""
        self.set_available(True, True)
        if self.in_RSSI != message.in_RSSI.value:
            self.in_RSSI = message.in_RSSI.value
            self.do_callback(SENSOR_RSSI_IN["id"])
        if self.out_RSSI != message.out_RSSI.value:
            self.out_RSSI = message.out_RSSI.value
            self.do_callback(SENSOR_RSSI_OUT["id"])
        if self.ping_ms != message.ping_ms.value:
            self.ping_ms = message.ping_ms.value
            self.do_callback(SENSOR_PING["id"])

    def _process_info_response(self, message):
        """ Process info response message"""
        self.stick.logger.debug("Response info message for node %s", self.get_mac())
        self.set_available(True)
        if message.relay_state.serialize() == b"01":
            if not self._relay_state:
                self._relay_state = True
                self.do_callback(SWITCH_RELAY["id"])
        else:
            if self._relay_state:
                self._relay_state = False
                self.do_callback(SWITCH_RELAY["id"])
        self._hardware_version = message.hw_ver.value.decode(UTF8_DECODE)
        self._firmware_version = message.fw_ver.value
        self._node_type = message.node_type.value
        self._last_info_message = message.timestamp
        if self._last_log_address != message.last_logaddr.value:
            self._last_log_address = message.last_logaddr.value
            self._last_log_collected = False
        self.stick.logger.debug("Node type        = %s", self.get_node_type())
        if not self.is_sed:
            self.stick.logger.debug("Relay state      = %s", str(self._relay_state))
        self.stick.logger.debug("Hardware version = %s", str(self._hardware_version))
        self.stick.logger.debug("Firmware version = %s", str(self._firmware_version))

    def _process_features_response(self, message):
        """ Process features message """
        self.stick.logger.info(
            "Node %s supports features %s", self.get_mac(), str(message.features.value)
        )
        self._features = message.features.value



"""
Use of this source code is governed by the MIT license found in the LICENSE file.

Plugwise switch node object
"""
from plugwise.constants import (
    HA_BINARY_SENSOR,
    HA_SENSOR,
    SENSOR_SWITCH,
)
from plugwise.nodes.sed import NodeSED
from plugwise.messages.responses import NodeSwitchGroupResponse


class PlugwiseSwitch(NodeSED):
    """provides interface to the Plugwise Switch nodes"""

    def __init__(self, mac, address, stick):
        super().__init__(mac, address, stick)
        self.categories = (HA_SENSOR, HA_BINARY_SENSOR)
        self._switch_state = False

    def get_switch_state(self):
        """ Return state of switch"""
        return self._switch_state

    def _on_SED_message(self, message):
        """
        Process received message
        """
        if isinstance(message, NodeSwitchGroupResponse):
            self.stick.logger.debug(
                "Switch group request %s received from %s for group id %s",
                str(message.power_state),
                self.get_mac(),
                str(message.group),
            )
            self._process_switch_group(message)
            self.stick.message_processed(message.seq_id)

    def _process_switch_group(self, message):
        """Switch group request from Scan"""
        if message.power_state == 0:
            # turn off => clear motion
            if self._switch_state:
                self._switch_state = False
                self.do_callback(SENSOR_SWITCH["id"])
        elif message.power_state == 1:
            # turn on => motion
            if not self._switch_state:
                self._switch_state = True
                self.do_callback(SENSOR_SWITCH["id"])
        else:
            self.stick.logger.debug(
                "Unknown power_state (%s) received from %s",
                str(message.power_state),
                self.get_mac(),
            )



"""
Use of this source code is governed by the MIT license found in the LICENSE file.

Plugwise Stealth node object
"""
from plugwise.node import PlugwiseNode
from plugwise.nodes.circle import PlugwiseCircle


class PlugwiseStealth(PlugwiseCircle):
    """provides interface to the Plugwise Stealth nodes"""

    def __init__(self, mac, address, stick):
        super().__init__(mac, address, stick)



"""
Use of this source code is governed by the MIT license found in the LICENSE file.

Main stick object to control associated plugwise plugs
"""
import logging
import time
import serial
import sys
import threading
from datetime import datetime, timedelta
from plugwise.constants import (
    ACCEPT_JOIN_REQUESTS,
    ACK_CLOCK_SET,
    ACK_ERROR,
    ACK_ACCEPT_JOINING_REQUEST,
    ACK_ON,
    ACK_OFF,
    ACK_SLEEP_SET,
    ACK_SUCCESS,
    ACK_REAL_TIME_CLOCK_SET,
    ACK_SCAN_PARAMETERS_SET,
    ACK_TIMEOUT,
    CB_JOIN_REQUEST,
    CB_NEW_NODE,
    MAX_TIME_DRIFT,
    MESSAGE_TIME_OUT,
    MESSAGE_RETRY,
    NACK_ON_OFF,
    NACK_REAL_TIME_CLOCK_SET,
    NACK_SCAN_PARAMETERS_SET,
    NACK_SLEEP_SET,
    NODE_TYPE_STICK,
    NODE_TYPE_CELSIUS_SED,
    NODE_TYPE_CELSIUS_NR,
    NODE_TYPE_CIRCLE_PLUS,
    NODE_TYPE_CIRCLE,
    NODE_TYPE_SWITCH,
    NODE_TYPE_SENSE,
    NODE_TYPE_SCAN,
    NODE_TYPE_STEALTH,
    SLEEP_TIME,
    WATCHDOG_DEAMON,
    UTF8_DECODE,
)
from plugwise.connections.socket import SocketConnection
from plugwise.connections.serial import PlugwiseUSBConnection
from plugwise.exceptions import (
    CirclePlusError,
    NetworkDown,
    PortError,
    StickInitError,
    TimeoutException,
)
from plugwise.message import PlugwiseMessage
from plugwise.messages.requests import (
    CircleClockGetRequest,
    CircleClockSetRequest,
    CirclePlusScanRequest,
    CircleCalibrationRequest,
    CirclePlusRealTimeClockGetRequest,
    CirclePlusRealTimeClockSetRequest,
    CirclePowerUsageRequest,
    CircleSwitchRelayRequest,
    NodeAllowJoiningRequest,
    NodeAddRequest,
    NodeInfoRequest,
    NodePingRequest,
    NodeRequest,
    NodeRemoveRequest,
    StickInitRequest,
)
from plugwise.messages.responses import (
    CircleCalibrationResponse,
    CircleClockResponse,
    CirclePlusRealTimeClockResponse,
    CirclePlusScanResponse,
    CirclePowerUsageResponse,
    NodeAckLargeResponse,
    NodeAckResponse,
    NodeAckSmallResponse,
    NodeAwakeResponse,
    NodeInfoResponse,
    NodeJoinAckResponse,
    NodeJoinAvailableResponse,
    NodePingResponse,
    NodeRemoveResponse,
    NodeResponse,
    StickInitResponse,
)
from plugwise.parser import PlugwiseParser
from plugwise.node import PlugwiseNode
from plugwise.nodes.circle import PlugwiseCircle
from plugwise.nodes.circle_plus import PlugwiseCirclePlus
from plugwise.nodes.sed import NodeSED
from plugwise.nodes.scan import PlugwiseScan
from plugwise.nodes.sense import PlugwiseSense
from plugwise.nodes.stealth import PlugwiseStealth
from plugwise.nodes.switch import PlugwiseSwitch
from plugwise.util import inc_seq_id, validate_mac
import queue


class stick(object):
    """
    Plugwise connection stick
    """

    def __init__(self, port, callback=None, print_progress=False):
        self.logger = logging.getLogger("python-plugwise")
        self._mac_stick = None
        self.port = port
        self.network_online = False
        self.circle_plus_mac = None
        self._circle_plus_discovered = False
        self._circle_plus_retries = 0
        self.network_id = None
        self.parser = PlugwiseParser(self)
        self._plugwise_nodes = {}
        self._nodes_registered = 0
        self._nodes_to_discover = {}
        self._nodes_not_discovered = {}
        self._nodes_off_line = 0
        self._discovery_finished = False
        self._messages_for_undiscovered_nodes = []
        self._accept_join_requests = ACCEPT_JOIN_REQUESTS
        self._stick_initialized = False
        self._stick_callbacks = {}
        self.last_ack_seq_id = None
        self.expected_responses = {}
        self.print_progress = print_progress
        self.timezone_delta = datetime.now().replace(
            minute=0, second=0, microsecond=0
        ) - datetime.utcnow().replace(minute=0, second=0, microsecond=0)
        self._run_receive_timeout_thread = False
        self._run_send_message_thread = False
        self._run_update_thread = False

        if callback:
            self.auto_initialize(callback)

    def auto_initialize(self, callback=None):
        """ automatic initialization """

        def init_finished():
            if not self.network_online:
                self.logger.Error("plugwise Zigbee network down")
            else:
                if self.print_progress:
                    print("Scan Plugwise network")
                self.scan(callback)

        try:
            if self.print_progress:
                print("Open port")
            self.connect()
            if self.print_progress:
                print("Initialize Plugwise USBstick")
            self.initialize_stick(init_finished)
        except PortError as e:
            self.logger.error("Failed to connect: '%s'", e)
        except StickInitError as e:
            self.logger.error("Failed to initialize USBstick: '%s'", e)
        except NetworkDown as e:
            self.logger.error("Failed to communicated: Plugwise Zigbee network")
        except TimeoutException as e:
            self.logger.error("Timeout exception while initializing USBstick")
        except Exception as e:
            self.logger.error("Unknown error : %s", e)

    def connect(self, callback=None):
        """ Connect to stick and raise error if it fails"""
        self.init_callback = callback
        # Open connection to USB Stick
        if ":" in self.port:
            self.logger.debug("Open socket connection to Plugwise Zigbee stick")
            self.connection = SocketConnection(self.port, self)
        else:
            self.logger.debug("Open USB serial connection to Plugwise Zigbee stick")
            self.connection = PlugwiseUSBConnection(self.port, self)
        self.connection.connect()

        self.logger.debug("Starting threads...")
        # receive timeout deamon
        self._run_receive_timeout_thread = True
        self._receive_timeout_thread = threading.Thread(
            None, self._receive_timeout_loop, "receive_timeout_thread", (), {}
        )
        self._receive_timeout_thread.daemon = True
        self._receive_timeout_thread.start()
        # send deamon
        self._send_message_queue = queue.Queue()
        self._run_send_message_thread = True
        self._send_message_thread = threading.Thread(
            None, self._send_message_loop, "send_messages_thread", (), {}
        )
        self._send_message_thread.daemon = True
        self._send_message_thread.start()
        # update deamon
        self._run_update_thread = False
        self._auto_update_timer = 0
        self._update_thread = threading.Thread(
            None, self._update_loop, "update_thread", (), {}
        )
        self._update_thread.daemon = True
        self.logger.debug("All threads started")

    def initialize_stick(self, callback=None, timeout=MESSAGE_TIME_OUT):
        # Initialize USBstick
        if not self.connection.is_connected():
            raise StickInitError

        def cb_stick_initialized():
            """ Callback when initialization of Plugwise USBstick is finished """
            self._stick_initialized = True

            # Start watchdog deamon
            self._run_watchdog = True
            self._watchdog_thread = threading.Thread(
                None, self._watchdog_loop, "watchdog_thread", (), {}
            )
            self._watchdog_thread.daemon = True
            self._watchdog_thread.start()

            # Try to discover Circle+
            if self.circle_plus_mac:
                self.discover_node(self.circle_plus_mac)
            if callback:
                callback()

        self.logger.debug("Send init request to Plugwise Zigbee stick")
        self.send(StickInitRequest(), cb_stick_initialized)
        time_counter = 0
        while not self._stick_initialized and (time_counter < timeout):
            time_counter += 0.1
            time.sleep(0.1)
        if not self._stick_initialized:
            raise StickInitError
        if not self.network_online:
            raise NetworkDown

    def initialize_circle_plus(self, callback=None, timeout=MESSAGE_TIME_OUT):
        # Initialize Circle+
        if (
            not self.connection.is_connected()
            or not self._stick_initialized
            or not self.circle_plus_mac
        ):
            raise StickInitError
        # discover circle+ node
        self.discover_node(self.circle_plus_mac)

        time_counter = 0
        while not self._circle_plus_discovered and (time_counter < timeout):
            time_counter += 0.1
            time.sleep(0.1)
        if not self._circle_plus_discovered:
            raise CirclePlusError

    def disconnect(self):
        """ Disconnect from stick and raise error if it fails"""
        self._run_watchdog = False
        self._run_update_thread = False
        self._auto_update_timer = 0
        self._run_send_message_thread = False
        self._run_receive_timeout_thread = False
        self.connection.disconnect()

    def subscribe_stick_callback(self, callback, callback_type):
        """ Subscribe callback to execute """
        if callback_type not in self._stick_callbacks:
            self._stick_callbacks[callback_type] = []
        self._stick_callbacks[callback_type].append(callback)

    def unsubscribe_stick_callback(self, callback, callback_type):
        """ Register callback to execute """
        if callback_type in self._stick_callbacks:
            self._stick_callbacks[callback_type].remove(callback)

    def do_callback(self, callback_type, callback_arg=None):
        """ Execute callbacks registered for specified callback type """
        if callback_type in self._stick_callbacks:
            for callback in self._stick_callbacks[callback_type]:
                try:
                    if callback_arg is None:
                        callback()
                    else:
                        callback(callback_arg)
                except Exception as e:
                    self.logger.error("Error while executing callback : %s", e)

    def _discover_after_scan(self):
        """ Helper to do callback for new node """
        node_discovered = None
        for mac in self._nodes_not_discovered.keys():
            if self._plugwise_nodes.get(mac, None):
                node_discovered = mac
                break
        if node_discovered:
            del self._nodes_not_discovered[node_discovered]
            self.do_callback(CB_NEW_NODE, node_discovered)

    def registered_nodes(self) -> int:
        """ Return number of nodes registered in Circle+ """
        # Include Circle+ too
        return self._nodes_registered + 1

    def nodes(self) -> list:
        """ Return list of mac addresses of discovered and supported plugwise nodes """
        return list(
            dict(
                filter(lambda item: item[1] is not None, self._plugwise_nodes.items())
            ).keys()
        )

    def node(self, mac: str) -> PlugwiseNode:
        """ Return specific Plugwise node object"""
        return self._plugwise_nodes.get(mac, None)

    def discover_node(self, mac: str, callback=None, force_discover=False) -> bool:
        """ Discovery of plugwise node """
        if validate_mac(mac) == True:
            if not self._plugwise_nodes.get(mac):
                if mac not in self._nodes_not_discovered.keys():
                    self._nodes_not_discovered[mac] = (
                        None,
                        None,
                    )
                    self.send(
                        NodeInfoRequest(bytes(mac, UTF8_DECODE)),
                        callback,
                    )
                else:
                    (firstrequest, lastrequest) = self._nodes_not_discovered[mac]
                    if not (firstrequest and lastrequest):
                        self.send(
                            NodeInfoRequest(bytes(mac, UTF8_DECODE)),
                            callback,
                        )
                    elif force_discover:
                        self.send(
                            NodeInfoRequest(bytes(mac, UTF8_DECODE)),
                            callback,
                        )
                return True
            else:
                return False
        else:
            return False

    def scan(self, callback=None):
        """ scan for connected plugwise nodes """

        def scan_finished(nodes_to_discover):
            """ Callback when scan is finished """
            time.sleep(1)
            self.logger.debug("Scan plugwise network finished")
            self._nodes_discovered = 0
            self._nodes_to_discover = nodes_to_discover
            self._nodes_registered = len(nodes_to_discover)
            self._discovery_finished = False

            def node_discovered(nodes_off_line=False):
                if nodes_off_line:
                    self._nodes_off_line += 1
                self._nodes_discovered += 1
                self.logger.debug(
                    "Discovered Plugwise node %s (%s off-line) of %s",
                    str(len(self._plugwise_nodes)),
                    str(self._nodes_off_line),
                    str(len(self._nodes_to_discover)),
                )
                if (len(self._plugwise_nodes) - 1 + self._nodes_off_line) >= len(
                    self._nodes_to_discover
                ):
                    if self._nodes_off_line == 0:
                        self._nodes_to_discover = {}
                        self._nodes_not_discovered = {}
                    else:
                        for mac in self._nodes_to_discover:
                            if mac not in self._plugwise_nodes.keys():
                                self.logger.info(
                                    "Failed to discover node type for registered MAC '%s'. This is expected for battery powered nodes, they will be discovered at their first awake",
                                    str(mac),
                                )
                            else:
                                if mac in self._nodes_not_discovered:
                                    del self._nodes_not_discovered[mac]
                    self._discovery_finished = True
                    if callback:
                        callback()

            def timeout_expired():
                if not self._discovery_finished:
                    for mac in self._nodes_to_discover:
                        if mac not in self._plugwise_nodes.keys():
                            self.logger.info(
                                "Failed to discover node type for registered MAC '%s'. This is expected for battery powered nodes, they will be discovered at their first awake",
                                str(mac),
                            )
                        else:
                            if mac in self._nodes_not_discovered:
                                del self._nodes_not_discovered[mac]
                    if callback:
                        callback()

            # setup timeout for loading nodes
            discover_timeout = (
                10 + (len(nodes_to_discover) * 2) + (MESSAGE_TIME_OUT * MESSAGE_RETRY)
            )
            self.discover_timeout = threading.Timer(
                discover_timeout, timeout_expired
            ).start()
            self.logger.debug("Start discovery of linked node types...")
            for mac in nodes_to_discover:
                self.discover_node(mac, node_discovered)

        def scan_circle_plus():
            """Callback when Circle+ is discovered"""
            if self._plugwise_nodes.get(self.circle_plus_mac):
                if self.print_progress:
                    print("Scan Circle+ for linked nodes")
                self.logger.debug("Scan Circle+ for linked nodes...")
                self._plugwise_nodes[self.circle_plus_mac].scan_for_nodes(scan_finished)
            else:
                self.logger.error(
                    "Circle+ is not discovered in %s", self._plugwise_nodes
                )

        # Discover Circle+
        if self.circle_plus_mac:
            if self._plugwise_nodes.get(self.circle_plus_mac):
                scan_circle_plus()
            else:
                if self.print_progress:
                    print("Discover Circle+")
                self.logger.debug("Discover Circle+ at %s", self.circle_plus_mac)
                self.discover_node(self.circle_plus_mac, scan_circle_plus)
        else:
            self.logger.error(
                "Plugwise stick not properly initialized, Circle+ MAC is missing."
            )

    def get_mac_stick(self) -> str:
        """Return mac address of USB-Stick"""
        if self._mac_stick:
            return self._mac_stick.decode(UTF8_DECODE)
        return None

    def allow_join_requests(self, enable: bool, accept: bool):
        """
        Enable or disable Plugwise network
        Automatically accept new join request
        """
        self.send(NodeAllowJoiningRequest(enable))
        if enable:
            self._accept_join_requests = accept
        else:
            self._accept_join_requests = False

    def node_join(self, mac: str, callback=None) -> bool:
        """Accept node to join Plugwise network by adding it in Circle+ memory"""
        if validate_mac(mac) == True:
            self.send(NodeAddRequest(bytes(mac, UTF8_DECODE), True), callback)
            return True
        else:
            self.logger.warning(
                "Invalid mac '%s' address, unable to join node manually.", mac
            )
        return False

    def node_unjoin(self, mac: str, callback=None) -> bool:
        """Remove node from the Plugwise network by deleting it from the Circle+ memory"""
        if validate_mac(mac) == True:
            self.send(
                NodeRemoveRequest(bytes(self.circle_plus_mac, UTF8_DECODE), mac),
                callback,
            )
            return True
        else:
            self.logger.warning(
                "Invalid mac '%s' address, unable to unjoin node manually.", mac
            )
        return False

    def _append_node(self, mac, address, node_type):
        """ Add Plugwise node to be controlled """
        self.logger.debug(
            "Add new node type (%s) with mac %s",
            str(node_type),
            mac,
        )
        if node_type == NODE_TYPE_CIRCLE_PLUS:
            if self.print_progress:
                print("Circle+ node found using mac " + mac)
            self._plugwise_nodes[mac] = PlugwiseCirclePlus(mac, address, self)
        elif node_type == NODE_TYPE_CIRCLE:
            if self.print_progress:
                print("Circle node found using mac " + mac)
            self._plugwise_nodes[mac] = PlugwiseCircle(mac, address, self)
        elif node_type == NODE_TYPE_SWITCH:
            if self.print_progress:
                print("Unsupported switch node found using mac " + mac)
            self._plugwise_nodes[mac] = None
        elif node_type == NODE_TYPE_SENSE:
            if self.print_progress:
                print("Sense node found using mac " + mac)
            self._plugwise_nodes[mac] = PlugwiseSense(mac, address, self)
        elif node_type == NODE_TYPE_SCAN:
            if self.print_progress:
                print("Scan node found using mac " + mac)
            self._plugwise_nodes[mac] = PlugwiseScan(mac, address, self)
        elif node_type == NODE_TYPE_CELSIUS_SED:
            if self.print_progress:
                print("Unsupported Celsius SED node found using mac " + mac)
            self._plugwise_nodes[mac] = None
        elif node_type == NODE_TYPE_CELSIUS_NR:
            if self.print_progress:
                print("Unsupported Celsius NR found using mac " + mac)
            self._plugwise_nodes[mac] = None
        elif node_type == NODE_TYPE_STEALTH:
            if self.print_progress:
                print("Stealth node found using mac " + mac)
            self._plugwise_nodes[mac] = PlugwiseStealth(mac, address, self)
        else:
            self.logger.warning("Unsupported node type '%s'", str(node_type))
            self._plugwise_nodes[mac] = None

        # process previous missed messages
        msg_to_process = self._messages_for_undiscovered_nodes[:]
        self._messages_for_undiscovered_nodes = []
        for msg in msg_to_process:
            self.new_message(msg)

    def _remove_node(self, mac):
        """
        remove circle from stick

        :return: None
        """
        if mac in self._plugwise_nodes:
            del self._plugwise_nodes[mac]

    def feed_parser(self, data):
        """ Feed parser with new data """
        assert isinstance(data, bytes)
        self.parser.feed(data)

    def send(self, request, callback=None, retry_counter=0):
        """
        Submit request message into Plugwise Zigbee network and queue expected response
        """
        assert isinstance(request, NodeRequest)
        if isinstance(request, CirclePowerUsageRequest):
            response_message = CirclePowerUsageResponse()
        elif isinstance(request, NodeInfoRequest):
            response_message = NodeInfoResponse()
        elif isinstance(request, NodePingRequest):
            response_message = NodePingResponse()
        elif isinstance(request, CircleSwitchRelayRequest):
            response_message = NodeAckLargeResponse()
        elif isinstance(request, CircleCalibrationRequest):
            response_message = CircleCalibrationResponse()
        elif isinstance(request, CirclePlusScanRequest):
            response_message = CirclePlusScanResponse()
        elif isinstance(request, CirclePlusRealTimeClockGetRequest):
            response_message = CirclePlusRealTimeClockResponse()
        elif isinstance(request, CircleClockGetRequest):
            response_message = CircleClockResponse()
        elif isinstance(request, StickInitRequest):
            response_message = StickInitResponse()
        else:
            response_message = None
        self._send_message_queue.put(
            [
                response_message,
                request,
                callback,
                retry_counter,
                None,
            ]
        )

    def _send_message_loop(self):
        """ deamon to send messages waiting in queue """
        while self._run_send_message_thread:
            try:
                request_set = self._send_message_queue.get(block=True, timeout=1)
            except queue.Empty:
                time.sleep(SLEEP_TIME)
            else:
                if self.last_ack_seq_id:
                    # Calc new seq_id based last received ack messsage
                    seq_id = inc_seq_id(self.last_ack_seq_id)
                else:
                    # first message, so use a fake seq_id
                    seq_id = b"0000"
                self.expected_responses[seq_id] = request_set
                if (
                    not isinstance(request_set[1], StickInitRequest)
                    and not isinstance(request_set[1], NodeAllowJoiningRequest)
                    and not isinstance(request_set[1], NodeAddRequest)
                ):
                    mac = request_set[1].mac.decode(UTF8_DECODE)
                    self.logger.info(
                        "send %s to %s using seq_id %s",
                        request_set[1].__class__.__name__,
                        mac,
                        str(seq_id),
                    )
                    if self._plugwise_nodes.get(mac):
                        self._plugwise_nodes[mac].last_request = datetime.now()
                    if self.expected_responses[seq_id][3] > 0:
                        self.logger.debug(
                            "Retry %s for message %s to %s",
                            str(self.expected_responses[seq_id][3]),
                            str(self.expected_responses[seq_id][1].__class__.__name__),
                            self.expected_responses[seq_id][1].mac.decode(UTF8_DECODE),
                        )
                else:
                    mac = ""
                    self.logger.info(
                        "send %s using seq_id %s",
                        request_set[1].__class__.__name__,
                        str(seq_id),
                    )
                self.expected_responses[seq_id][4] = datetime.now()
                self.connection.send(request_set[1])
                time.sleep(SLEEP_TIME)
                timeout_counter = 0
                # Wait max 1 second for acknowledge response
                while (
                    self.last_ack_seq_id != seq_id
                    and timeout_counter <= 10
                    and seq_id != b"0000"
                    and self.last_ack_seq_id != None
                ):
                    time.sleep(0.1)
                    timeout_counter += 1
                if timeout_counter > 10 and self._run_send_message_thread:
                    if seq_id in self.expected_responses:
                        if self.expected_responses[seq_id][3] <= MESSAGE_RETRY:
                            self.logger.info(
                                "Resend %s for %s because stick did not acknowledge request (%s), last seq_id=%s",
                                str(
                                    self.expected_responses[seq_id][
                                        1
                                    ].__class__.__name__
                                ),
                                mac,
                                str(seq_id),
                                str(self.last_ack_seq_id),
                            )
                            self.send(
                                self.expected_responses[seq_id][1],
                                self.expected_responses[seq_id][2],
                                self.expected_responses[seq_id][3] + 1,
                            )
                        else:
                            self.logger.info(
                                "Drop %s request with seq_id %s for mac %s because max (%s) retries reached, last seq_id=%s",
                                self.expected_responses[seq_id][1].__class__.__name__,
                                str(seq_id),
                                mac,
                                str(MESSAGE_RETRY),
                                str(self.last_ack_seq_id),
                            )
                        del self.expected_responses[seq_id]
        self.logger.debug("Send message loop stopped")

    def _receive_timeout_loop(self):
        """ deamon to time out requests without any (n)ack response message """
        while self._run_receive_timeout_thread:
            for seq_id in list(self.expected_responses.keys()):
                if self.expected_responses[seq_id][4] != None:
                    if self.expected_responses[seq_id][4] < (
                        datetime.now() - timedelta(seconds=MESSAGE_TIME_OUT)
                    ):
                        self.logger.debug(
                            "Timeout expired for message with sequence ID %s",
                            str(seq_id),
                        )
                        if self.expected_responses[seq_id][3] <= MESSAGE_RETRY:
                            self.logger.debug(
                                "Resend request %s",
                                str(
                                    self.expected_responses[seq_id][
                                        1
                                    ].__class__.__name__
                                ),
                            )
                            self.send(
                                self.expected_responses[seq_id][1],
                                self.expected_responses[seq_id][2],
                                self.expected_responses[seq_id][3] + 1,
                            )
                        else:
                            if isinstance(
                                self.expected_responses[seq_id][1], NodeAddRequest
                            ) or isinstance(
                                self.expected_responses[seq_id][1], StickInitRequest
                            ):
                                self.logger.info(
                                    "Drop %s request because max (%s) retries reached for seq id %s",
                                    self.expected_responses[seq_id][
                                        1
                                    ].__class__.__name__,
                                    str(MESSAGE_RETRY),
                                    str(seq_id),
                                )
                            else:
                                if self.expected_responses[seq_id][1].mac == "":
                                    mac = "<empty>"
                                else:
                                    mac = self.expected_responses[seq_id][1].mac.decode(
                                        UTF8_DECODE
                                    )
                                self.logger.info(
                                    "Drop %s request for mac %s because max (%s) retries reached for seq id %s",
                                    self.expected_responses[seq_id][
                                        1
                                    ].__class__.__name__,
                                    mac,
                                    str(MESSAGE_RETRY),
                                    str(seq_id),
                                )
                        del self.expected_responses[seq_id]
            receive_timeout_checker = 0
            while (
                receive_timeout_checker < MESSAGE_TIME_OUT
                and self._run_receive_timeout_thread
            ):
                time.sleep(1)
                receive_timeout_checker += 1
        self.logger.debug("Receive timeout loop stopped")

    def new_message(self, message: NodeResponse):
        """ Received message from Plugwise Zigbee network """

        # only save last seq_id and skip special ID's FFFD, FFFE, FFFF
        if self.last_ack_seq_id:
            if int(self.last_ack_seq_id, 16) < int(message.seq_id, 16) < 65533:
                self.last_ack_seq_id = message.seq_id
            elif message.seq_id == b"0000":
                self.last_ack_seq_id = b"0000"

        if not isinstance(message, NodeAckSmallResponse):
            mac = message.mac.decode(UTF8_DECODE)
            if not isinstance(message, NodeAckLargeResponse):
                self.logger.info(
                    "Received %s from %s with seq_id %s",
                    message.__class__.__name__,
                    mac,
                    str(message.seq_id),
                )

        if isinstance(message, NodeAckSmallResponse):
            if message.ack_id == ACK_SUCCESS:
                self.logger.debug(
                    "Received success response for request with sequence id %s",
                    str(message.seq_id),
                )
                self.message_processed(message.seq_id, message.ack_id, True)
            elif message.ack_id == ACK_TIMEOUT:
                self.logger.info(
                    "Received timeout response for request with sequence id %s",
                    str(message.seq_id),
                )
                self.message_processed(message.seq_id, message.ack_id, True)
            elif message.ack_id == ACK_ERROR:
                self.logger.info(
                    "Received error response for request with sequence id %s",
                    str(message.seq_id),
                )
                self.message_processed(message.seq_id, message.ack_id, True)
            else:
                if self.expected_responses.get(message.seq_id):
                    self.logger.info(
                        "Received unmanaged NodeAckSmallResponse %s message for request %s with sequence id %s",
                        str(message.ack_id),
                        str(
                            self.expected_responses[message.seq_id][
                                1
                            ].__class__.__name__
                        ),
                        str(message.seq_id),
                    )
                else:
                    self.logger.info(
                        "Received unmanaged NodeAckSmallResponse %s message for unknown request with sequence id %s",
                        str(message.ack_id),
                        str(message.seq_id),
                    )
        elif isinstance(message, NodeAckLargeResponse):
            if self._plugwise_nodes.get(mac):
                if message.ack_id == ACK_ON:
                    self.logger.info(
                        "Received relay switched on in response for CircleSwitchRelayRequest from %s with sequence id %s",
                        mac,
                        str(message.seq_id),
                    )
                    self._plugwise_nodes[mac].on_message(message)
                    self.message_processed(message.seq_id, message.ack_id)
                elif message.ack_id == ACK_OFF:
                    self.logger.info(
                        "Received relay switched off in response for CircleSwitchRelayRequest from %s with sequence id %s",
                        mac,
                        str(message.seq_id),
                    )
                    self._plugwise_nodes[mac].on_message(message)
                    self.message_processed(message.seq_id, message.ack_id)
                elif message.ack_id == NACK_ON_OFF:
                    self.logger.info(
                        "Received failed response for CircleSwitchRelayRequest from %s with sequence id %s",
                        mac,
                        str(message.seq_id),
                    )
                    self.message_processed(message.seq_id, message.ack_id)
                elif message.ack_id == ACK_SLEEP_SET:
                    self.logger.info(
                        "Received success sleep configuration response for NodeSleepConfigRequest from %s with sequence id %s",
                        mac,
                        str(message.seq_id),
                    )
                    self._plugwise_nodes[mac].on_message(message)
                    self.message_processed(message.seq_id, message.ack_id)
                elif message.ack_id == NACK_SLEEP_SET:
                    self.logger.warning(
                        "Received failed sleep configuration response for NodeSleepConfigRequest from %s with sequence id %s",
                        mac,
                        str(message.seq_id),
                    )
                    self._plugwise_nodes[mac].on_message(message)
                    self.message_processed(message.seq_id, message.ack_id)
                elif message.ack_id == ACK_ACCEPT_JOINING_REQUEST:
                    self.logger.info(
                        "Received success response for NodeAllowJoiningRequest from (circle+) %s with sequence id %s",
                        mac,
                        str(message.seq_id),
                    )
                    self.message_processed(message.seq_id, message.ack_id)
                elif message.ack_id == ACK_CLOCK_SET:
                    self.logger.info(
                        "Received success response for CircleClockSetRequest from %s with sequence id %s",
                        mac,
                        str(message.seq_id),
                    )
                    self.message_processed(message.seq_id, message.ack_id)
                else:
                    if self.expected_responses.get(message.seq_id):
                        self.logger.info(
                            "Received unmanaged NodeAckLargeResponse %s message from %s for request %s with sequence id %s",
                            str(message.ack_id),
                            mac,
                            str(
                                self.expected_responses[message.seq_id][
                                    1
                                ].__class__.__name__
                            ),
                            str(message.seq_id),
                        )
                    else:
                        self.logger.info(
                            "Received unmanaged NodeAckLargeResponse %s message from %s for unknown request with sequence id %s",
                            str(message.ack_id),
                            mac,
                            str(message.seq_id),
                        )
            else:
                self.logger.info(
                    "Received NodeAckLargeResponse %s message from unknown node %s with sequence id %s",
                    str(message.ack_id),
                    mac,
                    str(message.seq_id),
                )
        elif isinstance(message, NodeAckResponse):
            if self._plugwise_nodes.get(mac):
                if message.ack_id == ACK_SCAN_PARAMETERS_SET:
                    self.logger.info(
                        "Received success response for ScanConfigureRequest from %s with sequence id %s",
                        mac,
                        str(message.seq_id),
                    )
                    self._plugwise_nodes[mac].on_message(message)
                    self.message_processed(message.seq_id, message.ack_id)
                elif message.ack_id == NACK_SCAN_PARAMETERS_SET:
                    self.logger.info(
                        "Received failed response for ScanConfigureRequest from %s with sequence id %s",
                        mac,
                        str(message.seq_id),
                    )
                    self.message_processed(message.seq_id, message.ack_id)
                else:
                    if self.expected_responses.get(message.seq_id):
                        self.logger.info(
                            "Received unmanaged NodeAckResponse %s message from %s for request %s with sequence id %s",
                            str(message.ack_id),
                            mac,
                            str(
                                self.expected_responses[message.seq_id][
                                    1
                                ].__class__.__name__
                            ),
                            str(message.seq_id),
                        )
                    else:
                        self.logger.info(
                            "Received unmanaged NodeAckResponse %s message from %s for unknown request with sequence id %s",
                            str(message.ack_id),
                            mac,
                            str(message.seq_id),
                        )
            else:
                self.logger.info(
                    "Received NodeAckResponse %s message from unknown node %s with sequence id %s",
                    str(message.ack_id),
                    mac,
                    str(message.seq_id),
                )
        elif isinstance(message, StickInitResponse):
            self._mac_stick = message.mac
            if message.network_is_online.value == 1:
                self.network_online = True
            else:
                self.network_online = False
            # Replace first 2 charactors by 00 for mac of circle+ node
            self.circle_plus_mac = "00" + message.circle_plus_mac.value[2:].decode(
                UTF8_DECODE
            )
            self.network_id = message.network_id.value
            # The first StickInitResponse gives the actual sequence ID
            if b"0000" in self.expected_responses:
                seq_id = b"0000"
            else:
                seq_id = message.seq_id
            self.message_processed(seq_id)
        elif isinstance(message, NodeInfoResponse):
            self.logger.debug(
                "Received node info (%s) for NodeInfoRequest from %s with sequence id %s",
                str(message.node_type.value),
                mac,
                str(message.seq_id),
            )
            if not mac in self._plugwise_nodes:
                if message.node_type.value == NODE_TYPE_CIRCLE_PLUS:
                    self._circle_plus_discovered = True
                    self._append_node(mac, 0, message.node_type.value)
                    if mac in self._nodes_not_discovered:
                        del self._nodes_not_discovered[mac]
                else:
                    for mac_to_discover in self._nodes_to_discover:
                        if mac == mac_to_discover:
                            self._append_node(
                                mac,
                                self._nodes_to_discover[mac_to_discover],
                                message.node_type.value,
                            )
            if self._plugwise_nodes.get(mac):
                self._plugwise_nodes[mac].on_message(message)
                self.message_processed(message.seq_id)
        elif isinstance(message, NodeAwakeResponse):
            # Message from SED node notifying it is currently awake.
            # If node is not known do discovery first.
            self.logger.info(
                "Received NodeAwakeResponse message (%s) from %s with sequence id %s",
                str(message.awake_type.value),
                mac,
                str(message.seq_id),
            )
            if self._plugwise_nodes.get(mac):
                self._plugwise_nodes[mac].on_message(message)
            else:
                self.logger.info(
                    "Received NodeAwakeResponse message from unknown node with mac %s with sequence id %s, do discovery now",
                    mac,
                    str(message.seq_id),
                )
                self.discover_node(mac, self._discover_after_scan, True)
        elif isinstance(message, NodeJoinAvailableResponse):
            # Message from node that is not part of a plugwise network yet and wants to join
            self.logger.info(
                "Received NodeJoinAvailableResponse from node with mac %s",
                mac,
            )
            if not self._plugwise_nodes.get(mac):
                if self._accept_join_requests:
                    # Send accept join request
                    self.logger.info(
                        "Accepting network join request for node with mac %s",
                        mac,
                    )
                    self.send(NodeAddRequest(bytes(mac, UTF8_DECODE), True))
                    self.discover_node(mac, self._discover_after_scan)
                else:
                    self.logger.debug(
                        "New node with mac %s requesting to join Plugwise network, do callback",
                        mac,
                    )
                    self.do_callback(CB_JOIN_REQUEST, mac)
            else:
                self.logger.debug(
                    "Received node available message for node %s which is already joined.",
                    mac,
                )
        elif isinstance(message, NodeJoinAckResponse):
            # Notification mesage when node (re)joined existing network again.
            # Received when a SED (re)joins the network e.g. when you reinsert the battery of a Scan
            self.logger.info(
                "Received NodeJoinAckResponse from %s which has accepted or (re)joined this Plugwise network",
                mac,
            )
            if not self._plugwise_nodes.get(mac):
                self.discover_node(mac, self._discover_after_scan, True)
        elif isinstance(message, NodeRemoveResponse):
            # Conformation message a node is is removed from the Plugwise network
            unjoined_mac = message.node_mac_id.value
            if message.status.value == 1:
                if self._plugwise_nodes.get(unjoined_mac):
                    del self._plugwise_nodes[unjoined_mac]
                    self.logger.info(
                        "Received NodeRemoveResponse from node %s it has been unjoined from Plugwise network",
                        unjoined_mac,
                    )
                else:
                    self.logger.debug(
                        "Unknown node with mac %s has been unjoined from Plugwise network",
                        unjoined_mac,
                    )
            else:
                self.logger.warning(
                    "Node with mac %s failed to unjoin from Plugwise network ",
                    unjoined_mac,
                )
        else:
            if self._plugwise_nodes.get(mac):
                self._plugwise_nodes[mac].on_message(message)
                self.message_processed(message.seq_id)
            else:
                self.logger.info(
                    "Queue %s message because node with mac %s is not discovered yet.",
                    message.__class__.__name__,
                    mac,
                )
                self._messages_for_undiscovered_nodes.append(message)
                self.discover_node(mac)

    def message_processed(self, seq_id, ack_response=None, ack_small=False):
        """ Execute callback of received messages """
        do_callback = False
        do_resend = False
        if seq_id in self.expected_responses:
            self.logger.debug(
                "Process response to %s with seq id %s",
                self.expected_responses[seq_id][0].__class__.__name__,
                str(seq_id),
            )
            if self.expected_responses[seq_id][1].mac == "":
                mac = "<unknown>"
            else:
                mac = self.expected_responses[seq_id][1].mac.decode(UTF8_DECODE)

            if not ack_response:
                do_callback = True
            elif ack_response == ACK_SUCCESS:
                if ack_small:
                    self.logger.debug(
                        "Process small ACK_SUCCESS acknowledge for %s with seq_id %s",
                        str(self.expected_responses[seq_id][1].__class__.__name__),
                        str(seq_id),
                    )
                else:
                    self.logger.debug(
                        "Process large ACK_SUCCESS acknowledge for %s from %s with seq_id %s",
                        str(self.expected_responses[seq_id][1].__class__.__name__),
                        mac,
                        str(seq_id),
                    )
                    do_callback = True
            elif ack_response == ACK_TIMEOUT:
                self.logger.debug(
                    "Process ACK_TIMEOUT for %s with seq_id %s",
                    str(self.expected_responses[seq_id][1].__class__.__name__),
                    str(seq_id),
                )
                do_resend = True
            elif ack_response == ACK_ERROR:
                self.logger.debug(
                    "Process ACK_ERROR for %s with seq_id %s",
                    str(self.expected_responses[seq_id][1].__class__.__name__),
                    str(seq_id),
                )
                do_resend = True
            elif ack_response == ACK_ON:
                self.logger.debug(
                    "Process ACK_ON response for %s from %s with seq_id %s",
                    self.expected_responses[seq_id][0].__class__.__name__,
                    mac,
                    str(seq_id),
                )
                do_callback = True
            elif ack_response == ACK_OFF:
                self.logger.debug(
                    "Process ACK_OFF response for %s from %s with seq_id %s",
                    self.expected_responses[seq_id][0].__class__.__name__,
                    mac,
                    str(seq_id),
                )
                do_callback = True
            elif ack_response == ACK_ACCEPT_JOINING_REQUEST:
                self.logger.debug(
                    "Process ACK_ACCEPT_JOINING_REQUEST for %s from %s with seq_id %s",
                    str(self.expected_responses[seq_id][1].__class__.__name__),
                    mac,
                    str(seq_id),
                )
                do_callback = True
            elif ack_response == ACK_SLEEP_SET:
                self.logger.debug(
                    "Process ACK_SLEEP_SET for %s with seq_id %s",
                    str(self.expected_responses[seq_id][1].__class__.__name__),
                    str(seq_id),
                )
                do_callback = True
            elif ack_response == ACK_CLOCK_SET:
                self.logger.debug(
                    "Process ACK_CLOCK_SET for %s with seq_id %s",
                    str(self.expected_responses[seq_id][1].__class__.__name__),
                    str(seq_id),
                )
                do_callback = True
            elif ack_response == NACK_SLEEP_SET:
                self.logger.debug(
                    "Process NACK_SLEEP_SET for %s with seq_id %s",
                    str(self.expected_responses[seq_id][1].__class__.__name__),
                    str(seq_id),
                )
                do_resend = True
            elif ack_response == ACK_SCAN_PARAMETERS_SET:
                self.logger.debug(
                    "Process ACK_SCAN_PARAMETERS_SET for %s with seq_id %s",
                    str(self.expected_responses[seq_id][1].__class__.__name__),
                    str(seq_id),
                )
                do_callback = True
            elif ack_response == NACK_SCAN_PARAMETERS_SET:
                self.logger.debug(
                    "Process NACK_SCAN_PARAMETERS_SET for %s with seq_id %s",
                    str(self.expected_responses[seq_id][1].__class__.__name__),
                    str(seq_id),
                )
                do_resend = True
            elif ack_response == NACK_ON_OFF:
                self.logger.debug(
                    "Process NACK_ON_OFF for %s with seq_id %s",
                    str(self.expected_responses[seq_id][1].__class__.__name__),
                    str(seq_id),
                )
                do_resend = True
            elif ack_response == NACK_REAL_TIME_CLOCK_SET:
                self.logger.debug(
                    "Process NACK_REAL_TIME_CLOCK_SET for %s with seq_id %s",
                    str(self.expected_responses[seq_id][1].__class__.__name__),
                    str(seq_id),
                )
                do_resend = True
            else:
                self.logger.warning(
                    "Unknown ack_response %s for %s with seq_id %s",
                    str(ack_response),
                    str(self.expected_responses[seq_id][1].__class__.__name__),
                    str(seq_id),
                )

            if do_resend:
                if self.expected_responses[seq_id][3] <= MESSAGE_RETRY:
                    if (
                        isinstance(self.expected_responses[seq_id][1], NodeInfoRequest)
                        and not self._discovery_finished
                        and mac in self._nodes_not_discovered
                        and self.expected_responses[seq_id][2].__name__
                        == "node_discovered"
                    ):
                        # Time out for node which is not discovered yet
                        # to speedup the initial discover phase skip retries and mark node as not discovered.
                        self.logger.debug(
                            "Skip retries for %s to speedup discover process",
                            mac,
                        )
                        self.expected_responses[seq_id][2](True)
                    elif isinstance(
                        self.expected_responses[seq_id][1], NodeInfoRequest
                    ) or isinstance(
                        self.expected_responses[seq_id][1], NodePingRequest
                    ):
                        self.logger.info(
                            "Resend request %s for %s, retry %s of %s",
                            str(self.expected_responses[seq_id][1].__class__.__name__),
                            mac,
                            str(self.expected_responses[seq_id][3] + 1),
                            str(MESSAGE_RETRY + 1),
                        )
                        self.send(
                            self.expected_responses[seq_id][1],
                            self.expected_responses[seq_id][2],
                            self.expected_responses[seq_id][3] + 1,
                        )
                    else:
                        if (
                            self._plugwise_nodes.get(mac)
                            and self._plugwise_nodes[mac].get_available()
                        ):
                            self.logger.info(
                                "Resend request %s for %s, retry %s of %s",
                                str(
                                    self.expected_responses[seq_id][
                                        1
                                    ].__class__.__name__
                                ),
                                mac,
                                str(self.expected_responses[seq_id][3] + 1),
                                str(MESSAGE_RETRY + 1),
                            )
                            self.send(
                                self.expected_responses[seq_id][1],
                                self.expected_responses[seq_id][2],
                                self.expected_responses[seq_id][3] + 1,
                            )
                        else:
                            self.logger.debug(
                                "Do not resend request %s for %s, node is off-line",
                                str(
                                    self.expected_responses[seq_id][
                                        1
                                    ].__class__.__name__
                                ),
                                mac,
                            )
                else:
                    self.logger.info(
                        "Drop request for %s for %s because max retries %s reached",
                        str(self.expected_responses[seq_id][1].__class__.__name__),
                        mac,
                        str(MESSAGE_RETRY + 1),
                    )
                    if isinstance(
                        self.expected_responses[seq_id][1], NodeInfoRequest
                    ) or isinstance(
                        self.expected_responses[seq_id][1], NodePingRequest
                    ):
                        # Mark node as unavailable
                        if self._plugwise_nodes.get(mac):
                            if (
                                self._plugwise_nodes[mac].get_available()
                                and not self._plugwise_nodes[mac].is_sed()
                            ):
                                self.logger.info(
                                    "Mark %s as unavailabe because no response after %s retries",
                                    mac,
                                    str(MESSAGE_RETRY + 1),
                                )
                                self._plugwise_nodes[mac].set_available(False)
                    else:
                        # Failed request, do a ping to validate if node is available
                        if self._plugwise_nodes.get(mac):
                            if not self._plugwise_nodes[mac].is_sed():
                                self._plugwise_nodes[mac].ping()
                del self.expected_responses[seq_id]

            if do_callback:
                if self.expected_responses[seq_id][2]:
                    try:
                        self.expected_responses[seq_id][2]()
                    except Exception as e:
                        self.logger.error(
                            "Error while executing callback after processing message : %s",
                            e,
                        )
                del self.expected_responses[seq_id]

        else:
            if not self.last_ack_seq_id:
                if b"0000" in self.expected_responses:
                    self.expected_responses[seq_id] = self.expected_responses[b"0000"]
                    del self.expected_responses[b"0000"]
                self.last_ack_seq_id = seq_id
            else:
                self.logger.info(
                    "Response %s for unknown seq_id %s",
                    str(ack_response),
                    str(seq_id),
                )

    def _watchdog_loop(self):
        """
        Main worker loop to watch all other worker threads
        """
        time.sleep(5)
        circle_plus_retry_counter = 0
        while self._run_watchdog:
            # Connection
            if self.connection.is_connected():
                # Connection reader daemon
                if not self.connection.read_thread_alive():
                    self.logger.warning("Unexpected halt of connection reader thread")
                # Connection writer daemon
                if not self.connection.write_thread_alive():
                    self.logger.warning("Unexpected halt of connection writer thread")
            # receive timeout daemon
            if self._run_receive_timeout_thread:
                if not self._receive_timeout_thread.isAlive():
                    self.logger.warning(
                        "Unexpected halt of receive thread, restart thread",
                    )
                    self._receive_timeout_thread = threading.Thread(
                        None,
                        self._receive_timeout_loop,
                        "receive_timeout_thread",
                        (),
                        {},
                    )
                    self._receive_timeout_thread.daemon = True
                    self._receive_timeout_thread.start()
            # send message deamon
            if self._run_send_message_thread:
                if not self._send_message_thread.isAlive():
                    self.logger.warning(
                        "Unexpected halt of send thread, restart thread",
                    )
                    self._send_message_thread = threading.Thread(
                        None, self._send_message_loop, "send_messages_thread", (), {}
                    )
                    self._send_message_thread.daemon = True
                    self._send_message_thread.start()
            # Update daemon
            if self._run_update_thread:
                if not self._update_thread.isAlive():
                    self.logger.warning(
                        "Unexpected halt of update thread, restart thread",
                    )
                    self._run_update_thread = True
                    self._update_thread = threading.Thread(
                        None, self._update_loop, "update_thread", (), {}
                    )
                    self._update_thread.daemon = True
                    self._update_thread.start()
            # Circle+ discovery
            if self._circle_plus_discovered == False:
                # First hour every once an hour
                if self._circle_plus_retries < 60 or circle_plus_retry_counter > 60:
                    self.logger.info(
                        "Circle+ not yet discovered, resubmit discovery request",
                    )
                    self.discover_node(self.circle_plus_mac, self.scan)
                    self._circle_plus_retries += 1
                    circle_plus_retry_counter = 0
                circle_plus_retry_counter += 1
            watchdog_loop_checker = 0
            while watchdog_loop_checker < WATCHDOG_DEAMON and self._run_watchdog:
                time.sleep(1)
                watchdog_loop_checker += 1
        self.logger.debug("watchdog loop stopped")

    def _update_loop(self):
        """
        When node has not received any message during
        last 2 update polls, reset availability
        """
        self._run_update_thread = True
        self._auto_update_first_run = True
        day_of_month = datetime.now().day
        try:
            while self._run_update_thread:
                for mac in self._plugwise_nodes:
                    if self._plugwise_nodes[mac]:
                        # Check availability state of SED's
                        if self._plugwise_nodes[mac].is_sed():
                            if self._plugwise_nodes[mac].get_available():
                                if self._plugwise_nodes[mac].last_update < (
                                    datetime.now()
                                    - timedelta(
                                        minutes=(
                                            self._plugwise_nodes[
                                                mac
                                            ]._maintenance_interval
                                            + 1
                                        )
                                    )
                                ):
                                    self.logger.info(
                                        "No messages received within (%s minutes) of expected maintenance interval from node %s, mark as unavailable [%s > %s]",
                                        str(
                                            self._plugwise_nodes[
                                                mac
                                            ]._maintenance_interval
                                        ),
                                        mac,
                                        str(self._plugwise_nodes[mac].last_update),
                                        str(
                                            datetime.now()
                                            - timedelta(
                                                minutes=(
                                                    self._plugwise_nodes[
                                                        mac
                                                    ]._maintenance_interval
                                                    + 1
                                                )
                                            )
                                        ),
                                    )
                                    self._plugwise_nodes[mac].set_available(False)
                        else:
                            # Do ping request
                            self.logger.debug(
                                "Send ping to node %s",
                                mac,
                            )
                            self._plugwise_nodes[mac].ping()

                    # Only power use updates for supported nodes
                    if (
                        isinstance(self._plugwise_nodes[mac], PlugwiseCircle)
                        or isinstance(self._plugwise_nodes[mac], PlugwiseCirclePlus)
                        or isinstance(self._plugwise_nodes[mac], PlugwiseStealth)
                    ):
                        # Don't check at first time
                        self.logger.debug(
                            "Request current power usage for node %s", mac
                        )
                        if not self._auto_update_first_run and self._run_update_thread:
                            # Only request update if node is available
                            if self._plugwise_nodes[mac].get_available():
                                self.logger.debug(
                                    "Node '%s' is available for update request, last update (%s)",
                                    mac,
                                    str(self._plugwise_nodes[mac].get_last_update()),
                                )
                                # Skip update request if there is still an request expected to be received
                                open_requests_found = False
                                for seq_id in list(self.expected_responses.keys()):
                                    if isinstance(
                                        self.expected_responses[seq_id][1],
                                        CirclePowerUsageRequest,
                                    ):
                                        if mac == self.expected_responses[seq_id][
                                            1
                                        ].mac.decode(UTF8_DECODE):
                                            open_requests_found = True
                                            break
                                if not open_requests_found:
                                    self._plugwise_nodes[mac].update_power_usage()
                                # Refresh node info once per hour and request power use afterwards
                                if self._plugwise_nodes[mac]._last_info_message != None:
                                    if self._plugwise_nodes[mac]._last_info_message < (
                                        datetime.now().replace(
                                            minute=0,
                                            second=0,
                                            microsecond=0,
                                        )
                                    ):
                                        self._plugwise_nodes[mac]._request_info(
                                            self._plugwise_nodes[
                                                mac
                                            ]._request_power_buffer
                                        )
                                if not self._plugwise_nodes[mac]._last_log_collected:
                                    self._plugwise_nodes[mac]._request_power_buffer()
                        else:
                            if self._run_update_thread:
                                self.logger.debug(
                                    "First request for current power usage for node %s",
                                    mac,
                                )
                                self._plugwise_nodes[mac].update_power_usage()
                        self._auto_update_first_run = False

                        # Sync internal clock of all available Circle and Circle+ nodes once a day
                        if datetime.now().day != day_of_month:
                            day_of_month = datetime.now().day
                            if self._plugwise_nodes[mac].get_available():
                                self._plugwise_nodes[mac].sync_clock()

                # Try to rediscover node(s) which where not available at initial scan
                # Do this the first hour at every update, there after only once an hour
                for mac in self._nodes_not_discovered:
                    (firstrequest, lastrequest) = self._nodes_not_discovered[mac]
                    if firstrequest and lastrequest:
                        if (firstrequest + timedelta(hours=1)) > datetime.now():
                            # first hour, so do every update a request
                            self.logger.debug(
                                "Try rediscovery of node %s",
                                mac,
                            )
                            self.discover_node(mac, self._discover_after_scan, True)
                            self._nodes_not_discovered[mac] = (
                                firstrequest,
                                datetime.now(),
                            )
                        else:
                            if (lastrequest + timedelta(hours=1)) < datetime.now():
                                self.logger.debug(
                                    "Try rediscovery of node %s",
                                    mac,
                                )
                                self.discover_node(mac, self._discover_after_scan, True)
                                self._nodes_not_discovered[mac] = (
                                    firstrequest,
                                    datetime.now(),
                                )
                    else:
                        self.logger.debug(
                            "Try rediscovery of node %s",
                            mac,
                        )
                        self.discover_node(mac, self._discover_after_scan, True)
                        self._nodes_not_discovered[mac] = (
                            datetime.now(),
                            datetime.now(),
                        )
                if self._auto_update_timer and self._run_update_thread:
                    update_loop_checker = 0
                    while (
                        update_loop_checker < self._auto_update_timer
                        and self._run_update_thread
                    ):
                        time.sleep(1)
                        update_loop_checker += 1

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            self.logger.error(
                "Error at line %s of _update_loop : %s", exc_tb.tb_lineno, e
            )
        self.logger.debug("Update loop stopped")

    def auto_update(self, timer=None):
        """
        setup auto update polling for power usage.
        """
        if timer == 0:
            self._run_update_thread = False
            self._auto_update_timer = 0
        else:
            self._auto_update_timer = 5
            if timer == None:
                # Timer based on number of nodes and 3 seconds per node
                self._auto_update_timer = len(self._plugwise_nodes) * 3
            elif timer > 5:
                self._auto_update_timer = timer
            if not self._run_update_thread:
                self._update_thread.start()



"""
Use of this source code is governed by the MIT license found in the LICENSE file.

Base for serial or socket connections
"""
from plugwise.constants import SLEEP_TIME
from plugwise.message import PlugwiseMessage
import queue
import threading
import time


class StickConnection(object):
    """ Generic Plugwise stick connection"""

    def __init__(self, port, stick=None):
        self.port = port
        self.stick = stick
        self.run_reader_thread = False
        self.run_writer_thread = False
        self._is_connected = False
        self._writer = None

    ################################################
    ###             Open connection              ###
    ################################################

    def connect(self):
        """Open the connection"""
        if not self._is_connected:
            self._open_connection()

    def _open_connection(self):
        """Placeholder to initialize the connection"""
        raise NotImplementedError

    ################################################
    ###                   Reader                 ###
    ################################################

    def _reader_start(self, name):
        """Start the reader thread to receive data"""
        self._reader_thread = threading.Thread(None, self._reader_deamon, name, (), {})
        self.run_reader_thread = True
        self._reader_thread.start()

    def _reader_deamon(self):
        """Thread to collect available data from connection"""
        while self.run_reader_thread:
            data = self._read_data()
            if data:
                self.stick.feed_parser(data)
            time.sleep(0.01)
        self.stick.logger.debug("Reader deamon stopped")

    def _read_data(self):
        """Placeholder to receive message from the connection"""
        raise NotImplementedError

    ################################################
    ###                 Writer                   ###
    ################################################

    def _writer_start(self, name: str):
        """Start the writer thread to send data"""
        self._write_queue = queue.Queue()
        self._writer_thread = threading.Thread(None, self._writer_deamon, name, (), {})
        self._writer_thread.daemon = True
        self.run_writer_thread = True
        self._writer_thread.start()

    def _writer_deamon(self):
        """Thread to write data from queue to existing connection."""
        while self.run_writer_thread:
            try:
                (message, callback) = self._write_queue.get(block=True, timeout=1)
            except queue.Empty:
                time.sleep(SLEEP_TIME)
            else:
                self.stick.logger.debug(
                    "Sending %s to plugwise stick (%s)",
                    message.__class__.__name__,
                    message.serialize(),
                )
                self._write_data(message.serialize())
                time.sleep(SLEEP_TIME)
                if callback:
                    callback()
        self.stick.logger.debug("Writer deamon stopped")

    def _write_data(self, data):
        """Placeholder to write message to the connection"""
        raise NotImplementedError

    def send(self, message: PlugwiseMessage, callback=None):
        """Add message to write queue."""
        self._write_queue.put_nowait((message, callback))

    ################################################
    ###             Connection state             ###
    ################################################

    def is_connected(self):
        """Return connection state"""
        return self._is_connected

    def read_thread_alive(self):
        """Return state of write thread"""
        return self._reader_thread.isAlive() if self.run_reader_thread else False

    def write_thread_alive(self):
        """Return state of write thread"""
        return self._writer_thread.isAlive() if self.run_writer_thread else False

    ################################################
    ###             Close connection             ###
    ################################################

    def disconnect(self):
        """Close the connection"""
        if self._is_connected:
            self._is_connected = False
            self.run_writer_thread = False
            self.run_reader_thread = False
            max_wait = 5 * SLEEP_TIME
            while self._writer_thread.isAlive():
                time.sleep(SLEEP_TIME)
                max_wait -= SLEEP_TIME
            self._close_connection()

    def _close_connection(self):
        """Placeholder to close the port"""
        raise NotImplementedError



"""
Use of this source code is governed by the MIT license found in the LICENSE file.

Serial connection
"""
import serial
from plugwise.constants import (
    BAUD_RATE,
    BYTE_SIZE,
    PARITY,
    SLEEP_TIME,
    STOPBITS,
)
from plugwise.connections.connection import StickConnection
from plugwise.exceptions import PortError
from plugwise.message import PlugwiseMessage


class PlugwiseUSBConnection(StickConnection):
    """simple wrapper around serial module"""

    def __init__(self, port, stick=None):
        super().__init__(port, stick)
        self._baud = BAUD_RATE
        self._byte_size = BYTE_SIZE
        self._stopbits = STOPBITS
        self._parity = serial.PARITY_NONE

    def _open_connection(self):
        """Open serial port"""
        self.stick.logger.debug("Open serial port %s", self.port)
        try:
            self._serial = serial.Serial(
                port=self.port,
                baudrate=self._baud,
                bytesize=self._byte_size,
                parity=self._parity,
                stopbits=self._stopbits,
                timeout=1,
            )
        except serial.serialutil.SerialException as err:
            self.stick.logger.debug(
                "Failed to connect to serial port %s, %s",
                self.port,
                err,
            )
            raise PortError(err)
        self._is_connected = self._serial.isOpen()
        if self._is_connected:
            self._reader_start("serial_reader_thread")
            self._writer_start("serial_writer_thread")
            self.stick.logger.debug(
                "Successfully connected to serial port %s", self.port
            )
        else:
            self.stick.logger.error(
                "Failed to open serial port %s",
                self.port,
            )

    def _close_connection(self):
        """Close serial port."""
        try:
            self._serial.close()
        except serial.serialutil.SerialException as err:
            self.stick.logger.debug(
                "Failed to close serial port %s, %s",
                self.port,
                err,
            )
            raise PortError(err)

    def _read_data(self):
        """Read thread."""
        if self._is_connected:
            try:
                serial_data = self._serial.read_all()
            except serial.serialutil.SerialException as err:
                self.stick.logger.debug(
                    "Error while reading data from serial port : %s", err
                )
                self._is_connected = False
                raise PortError(err)
            except Exception as e:
                self.stick.logger.debug("Error _read_data : %s", err)
            return serial_data
        return None

    def _write_data(self, data):
        """Write data to serial port"""
        try:
            self._serial.write(data)
        except serial.serialutil.SerialException as err:
            self.stick.logger.debug("Error while writing data to serial port : %s", err)
            self._is_connected = False
            raise PortError(err)



"""
Use of this source code is governed by the MIT license found in the LICENSE file.

Base for Plugwise messages
"""

from plugwise.constants import (
    MESSAGE_FOOTER,
    MESSAGE_HEADER,
    UTF8_DECODE,
)
from plugwise.util import crc_fun


class ParserError(Exception):
    """
    Error when invalid message is received
    """

    pass


class PlugwiseMessage(object):
    def serialize(self):
        """
        return message in a serialized format that can be sent out on wire

        return: bytes
        """
        args = b"".join(a.serialize() for a in self.args)
        msg = self.ID
        if self.mac != "":
            msg += self.mac
        msg += args
        checksum = self.calculate_checksum(msg)
        return MESSAGE_HEADER + msg + checksum + MESSAGE_FOOTER

    def calculate_checksum(self, s):
        return bytes("%04X" % crc_fun(s), UTF8_DECODE)



"""
Use of this source code is governed by the MIT license found in the LICENSE file.

Plugwise Scan node object
"""
from plugwise.constants import (
    ACK_SCAN_PARAMETERS_SET,
    HA_BINARY_SENSOR,
    HA_SENSOR,
    NACK_SCAN_PARAMETERS_SET,
    SCAN_DAYLIGHT_MODE,
    SCAN_SENSITIVITY_HIGH,
    SCAN_SENSITIVITY_MEDIUM,
    SCAN_SENSITIVITY_OFF,
    SCAN_MOTION_RESET_TIMER,
    SCAN_SENSITIVITY,
    SENSOR_AVAILABLE,
    SENSOR_PING,
    SENSOR_RSSI_IN,
    SENSOR_RSSI_OUT,
    SENSOR_MOTION,
)
from plugwise.nodes.sed import NodeSED
from plugwise.message import PlugwiseMessage
from plugwise.messages.responses import NodeAckResponse, NodeSwitchGroupResponse
from plugwise.messages.requests import (
    ScanConfigureRequest,
    ScanLightCalibrateRequest,
)


class PlugwiseScan(NodeSED):
    """provides interface to the Plugwise Scan nodes"""

    def __init__(self, mac, address, stick):
        super().__init__(mac, address, stick)
        self.categories = (HA_SENSOR, HA_BINARY_SENSOR)
        self.sensors = (
            SENSOR_AVAILABLE["id"],
            SENSOR_PING["id"],
            SENSOR_MOTION["id"],
            SENSOR_RSSI_IN["id"],
            SENSOR_RSSI_OUT["id"],
        )
        self._motion_state = False
        self._motion_reset_timer = None
        self._daylight_mode = None
        self._sensitivity = None
        self._new_motion_reset_timer = None
        self._new_daylight_mode = None
        self._new_sensitivity = None

    def get_motion(self) -> bool:
        """ Return motion state"""
        return self._motion_state

    def _on_SED_message(self, message):
        """
        Process received message
        """
        if isinstance(message, NodeSwitchGroupResponse):
            self.stick.logger.debug(
                "Switch group %s to state %s received from %s",
                str(message.group.value),
                str(message.power_state.value),
                self.get_mac(),
            )
            self._process_switch_group(message)
        elif isinstance(message, NodeAckResponse):
            self._process_ack_message(message)
        else:
            self.stick.logger.info(
                "Unsupported message %s received from %s",
                message.__class__.__name__,
                self.get_mac(),
            )

    def _process_ack_message(self, message):
        """Process acknowledge message"""
        if message.ack_id == ACK_SCAN_PARAMETERS_SET:
            self._motion_reset_timer = self._new_motion_reset_timer
            self._daylight_mode = self._new_daylight_mode
            self._sensitivity = self._new_sensitivity
        else:
            self.stick.logger.info(
                "Unsupported ack message %s received for %s",
                str(message.ack_id),
                self.get_mac(),
            )

    def _process_switch_group(self, message):
        """Switch group request from Scan"""
        if message.power_state.value == 0:
            # turn off => clear motion
            if self._motion_state:
                self._motion_state = False
                self.do_callback(SENSOR_MOTION["id"])
        elif message.power_state.value == 1:
            # turn on => motion
            if not self._motion_state:
                self._motion_state = True
                self.do_callback(SENSOR_MOTION["id"])
        else:
            self.stick.logger.warning(
                "Unknown power_state (%s) received from %s",
                str(message.power_state.value),
                self.get_mac(),
            )

    def CalibrateLight(self, callback=None):
        """Queue request to calibration light sensitivity"""
        self._queue_request(ScanLightCalibrateRequest(self.mac), callback)

    def Configure_scan(
        self,
        motion_reset_timer=SCAN_MOTION_RESET_TIMER,
        sensitivity_level=SCAN_SENSITIVITY_MEDIUM,
        daylight_mode=SCAN_DAYLIGHT_MODE,
        callback=None,
    ):
        """Queue request to set motion reporting settings"""
        self._new_motion_reset_timer = motion_reset_timer
        self._new_daylight_mode = daylight_mode
        if sensitivity_level == SCAN_SENSITIVITY_HIGH:
            sensitivity_value = 20  # b'14'
        elif sensitivity_level == SCAN_SENSITIVITY_MEDIUM:
            sensitivity_value = 30  # b'1E'
        elif sensitivity_level == SCAN_SENSITIVITY_OFF:
            sensitivity_value = 255  # b'FF'
        self._new_sensitivity = sensitivity_level
        self._queue_request(
            ScanConfigureRequest(
                self.mac, motion_reset_timer, sensitivity_value, daylight_mode
            ),
            callback,
        )

    def SetMotionAction(self, callback=None):
        """Queue Configure Scan to signal motion"""
        # TODO:

        # self._queue_request(NodeSwitchGroupRequest(self.mac), callback)



# Use of this source code is governed by the MIT license found in the LICENSE file.

import logging
from plugwise.constants import (
    MESSAGE_FOOTER,
    MESSAGE_HEADER,
)
from plugwise.message import PlugwiseMessage
from plugwise.messages.responses import (
    CircleCalibrationResponse,  # 0027
    CirclePlusConnectResponse,  # 0005
    CirclePlusQueryEndResponse,  # 0003
    CirclePlusQueryResponse,  # 0002
    CirclePlusRealTimeClockResponse,  # 003A
    CirclePlusScanResponse,  # 0019
    CirclePowerBufferResponse,  # 0049
    CirclePowerUsageResponse,  # 0013
    CircleClockResponse,  # 003F
    NodeAckLargeResponse,  # 0000
    NodeAckResponse,  # 0100
    NodeAckSmallResponse,  # 0000
    NodeFeaturesResponse,  # 0060
    NodeInfoResponse,  # 0024
    NodeJoinAvailableResponse,  # 0006
    NodeJoinAckResponse,  # 0061
    NodePingResponse,  # 000E
    NodeSwitchGroupResponse,  # 0056
    NodeRemoveResponse,  # 001D
    NodeAwakeResponse,  # 004F
    SenseReportResponse,  # 0105
    StickInitResponse,  # 0011
)


class PlugwiseParser(object):
    """
    Transform Plugwise message from wire format to response message object
    """

    def __init__(self, stick):
        self.stick = stick
        self._buffer = bytes([])
        self._parsing = False
        self._message = None

    def feed(self, data):
        """
        Add new incoming data to buffer and try to process
        """
        self.stick.logger.debug("Feed data: %s", str(data))
        self._buffer += data
        if len(self._buffer) >= 8:
            if not self._parsing:
                self.parse_data()

    def next_message(self, message):
        """
        Process next packet if present
        """
        try:
            self.stick.new_message(message)
        except Exception as e:
            self.stick.logger.error(
                "Error while processing %s message : %s",
                self._message.__class__.__name__,
                e,
            )

    def parse_data(self):
        """
        Process next set of packet data
        """
        self.stick.logger.debug("Parse data: %s ", str(self._buffer))
        if self._parsing == False:
            self._parsing = True

            # Lookup header of message in buffer
            self.stick.logger.debug(
                "Lookup message header (%s) in (%s)",
                str(MESSAGE_HEADER),
                str(self._buffer),
            )
            header_index = self._buffer.find(MESSAGE_HEADER)
            if header_index == -1:
                self.stick.logger.debug("No valid message header found yet")
            else:
                self.stick.logger.debug(
                    "Valid message header found at index %s", str(header_index)
                )
                self._buffer = self._buffer[header_index:]

                # Header available, lookup footer of message in buffer
                self.stick.logger.debug(
                    "Lookup message footer (%s) in (%s)",
                    str(MESSAGE_FOOTER),
                    str(self._buffer),
                )
                footer_index = self._buffer.find(MESSAGE_FOOTER)
                if footer_index == -1:
                    self.stick.logger.debug("No valid message footer found yet")
                else:
                    self.stick.logger.debug(
                        "Valid message footer found at index %s", str(footer_index)
                    )
                    seq_id = self._buffer[8:12]
                    # First check for known sequence ID's
                    if seq_id == b"FFFD":
                        self._message = NodeJoinAckResponse()
                    elif seq_id == b"FFFE":
                        self._message = NodeAwakeResponse()
                    elif seq_id == b"FFFF":
                        self._message = NodeSwitchGroupResponse()
                    else:
                        # No fixed sequence ID, Continue at message ID
                        message_id = self._buffer[4:8]
                        if message_id == b"0000":
                            if footer_index == 20:
                                # Short acknowledge message without MAC
                                self._message = NodeAckSmallResponse()
                            elif footer_index == 36:
                                # Large ack message with a MAC
                                self._message = NodeAckLargeResponse()
                            else:
                                self.stick.logger.error(
                                    "Skip unknown ACK message size of %s : %s",
                                    str(footer_index + 2),
                                    str(self._buffer[: footer_index + 2]),
                                )
                        elif message_id == b"0002":
                            self._message = CirclePlusQueryResponse()
                        elif message_id == b"0003":
                            self._message = CirclePlusQueryEndResponse()
                        elif message_id == b"0005":
                            self._message = CirclePlusConnectResponse()
                        elif message_id == b"0006":
                            self._message = NodeJoinAvailableResponse()
                        elif message_id == b"000E":
                            self._message = NodePingResponse()
                        elif message_id == b"0011":
                            self._message = StickInitResponse()
                        elif message_id == b"0013":
                            self._message = CirclePowerUsageResponse()
                        elif message_id == b"0019":
                            self._message = CirclePlusScanResponse()
                        elif message_id == b"001D":
                            self._message = NodeRemoveResponse()
                        elif message_id == b"0024":
                            self._message = NodeInfoResponse()
                        elif message_id == b"0027":
                            self._message = CircleCalibrationResponse()
                        elif message_id == b"003A":
                            self._message = CirclePlusRealTimeClockResponse()
                        elif message_id == b"003F":
                            self._message = CircleClockResponse()
                        elif message_id == b"0049":
                            self._message = CirclePowerBufferResponse()
                        elif message_id == b"0060":
                            self._message = NodeFeaturesResponse()
                        elif message_id == b"0100":
                            self._message = NodeAckResponse()
                        elif message_id == b"0105":
                            self._message = SenseReportResponse()
                        elif footer_index < 28:
                            self.stick.logger.error(
                                "Received message %s to small, skip parsing",
                                self._buffer[: footer_index + 2],
                            )
                        else:
                            # Lookup expected message based on request
                            self.stick.logger.info(
                                "Unknown message received, id=%s, data=%s",
                                str(message_id),
                                str(self._buffer[: footer_index + 2]),
                            )
                            if message_id != b"0000":
                                self.stick.logger.debug(
                                    "Message id %s",
                                    str(message_id),
                                )
                            if seq_id in self.stick.expected_responses:
                                self._message = self.stick.expected_responses[seq_id][0]
                                self.stick.logger.debug(
                                    "Expected %s for message id %s",
                                    self._message.__class__.__name__,
                                    str(message_id),
                                )
                            else:
                                self.stick.logger.debug(
                                    "No expected message type found for sequence id %s in %s",
                                    str(seq_id),
                                    self.stick.expected_responses.keys(),
                                )
                                self.stick.logger.debug(
                                    "Message %s",
                                    str(self._buffer[: footer_index + 2]),
                                )
                    # Decode message
                    if isinstance(self._message, PlugwiseMessage):
                        if len(self._buffer[: footer_index + 2]) == len(self._message):
                            valid_message = False
                            try:
                                self._message.deserialize(
                                    self._buffer[: footer_index + 2]
                                )
                                valid_message = True
                            except Exception as e:
                                self.stick.logger.error(
                                    "Error while decoding received %s message (%s)",
                                    self._message.__class__.__name__,
                                    str(self._buffer[: footer_index + 2]),
                                )
                                self.stick.logger.error(e)
                            # Submit message
                            if valid_message:
                                self.next_message(self._message)
                        else:
                            self.stick.logger.error(
                                "Skip message, received %s bytes of expected %s bytes for message %s",
                                len(self._buffer[: footer_index + 2]),
                                len(self._message),
                                self._message.__class__.__name__,
                            )
                        # Parse remaining buffer
                        self.reset_parser(self._buffer[footer_index + 2 :])
                    else:
                        # skip this message, so remove header from buffer
                        self.reset_parser(self._buffer[6:])
            self._parsing = False
        else:
            self.stick.logger.debug("Skip parsing session")

    def reset_parser(self, new_buffer=bytes([])):
        self.stick.logger.debug("Reset parser : %s", new_buffer)
        if new_buffer == b"\x83":
            # Skip additional byte sometimes appended after footer
            self._buffer = bytes([])
        else:
            self._buffer = new_buffer
        self._message = None
        self._parsing = False
        if len(self._buffer) > 0:
            self.parse_data()



"""
Use of this source code is governed by the MIT license found in the LICENSE file.

Plugwise Sense node object
"""
from plugwise.constants import (
    HA_BINARY_SENSOR,
    HA_SENSOR,
    SENSE_HUMIDITY_MULTIPLIER,
    SENSE_HUMIDITY_OFFSET,
    SENSE_TEMPERATURE_MULTIPLIER,
    SENSE_TEMPERATURE_OFFSET,
    SENSOR_AVAILABLE,
    SENSOR_HUMIDITY,
    SENSOR_RSSI_IN,
    SENSOR_RSSI_OUT,
    SENSOR_TEMPERATURE,
)
from plugwise.nodes.sed import NodeSED
from plugwise.message import PlugwiseMessage
from plugwise.messages.responses import SenseReportResponse


class PlugwiseSense(NodeSED):
    """provides interface to the Plugwise Sense nodes"""

    def __init__(self, mac, address, stick):
        super().__init__(mac, address, stick)
        self.categories = (HA_SENSOR, HA_BINARY_SENSOR)
        self.sensors = (
            SENSOR_AVAILABLE["id"],
            SENSOR_TEMPERATURE["id"],
            SENSOR_HUMIDITY["id"],
            SENSOR_RSSI_IN["id"],
            SENSOR_RSSI_OUT["id"],
        )
        self._temperature = None
        self._humidity = None

    def get_temperature(self) -> int:
        """ Return the current temperature """
        return self._temperature

    def get_humidity(self) -> int:
        """ Return the current humidity """
        return self._humidity

    def _on_SED_message(self, message):
        """
        Process received message
        """
        if isinstance(message, SenseReportResponse):
            self._process_sense_report(message)
        else:
            self.stick.logger.info(
                "Unsupported message %s received from %s",
                message.__class__.__name__,
                self.get_mac(),
            )

    def _process_sense_report(self, message):
        """ process sense report message to extract current temperature and humidity values """
        if message.temperature.value != 65535:
            new_temperature = int(
                SENSE_TEMPERATURE_MULTIPLIER * (message.temperature.value / 65536)
                - SENSE_TEMPERATURE_OFFSET
            )
            if self._temperature != new_temperature:
                self._temperature = new_temperature
                self.stick.logger.debug(
                    "Sense report received from %s with new temperature level of %s",
                    self.get_mac(),
                    str(self._temperature),
                )
                self.do_callback(SENSOR_TEMPERATURE["id"])
        if message.humidity.value != 65535:
            new_humidity = int(
                SENSE_HUMIDITY_MULTIPLIER * (message.humidity.value / 65536)
                - SENSE_HUMIDITY_OFFSET
            )
            if self._humidity != new_humidity:
                self._humidity = new_humidity
                self.stick.logger.debug(
                    "Sense report received from %s with new humidity level of %s",
                    self.get_mac(),
                    str(self._humidity),
                )
                self.do_callback(SENSOR_HUMIDITY["id"])



"""
Use of this source code is governed by the MIT license found in the LICENSE file.

All known response messages to be received from plugwise devices
"""
from datetime import datetime
from plugwise.constants import (
    MESSAGE_FOOTER,
    MESSAGE_HEADER,
    MESSAGE_LARGE,
    MESSAGE_SMALL,
)
from plugwise.exceptions import ProtocolError
from plugwise.message import PlugwiseMessage
from plugwise.util import (
    DateTime,
    Float,
    Int,
    LogAddr,
    RealClockDate,
    RealClockTime,
    String,
    UnixTimestamp,
    Time,
)


class NodeResponse(PlugwiseMessage):
    """
    Base class for response messages received by USB-Stick.
    """

    def __init__(self, format_size=None):
        super().__init__()
        self.format_size = format_size
        self.params = []
        self.mac = None
        self.timestamp = datetime.now()
        self.seq_id = None
        self.msg_id = None
        self.ack_id = None
        if self.format_size == MESSAGE_SMALL:
            self.len_correction = -12
        elif self.format_size == MESSAGE_LARGE:
            self.len_correction = 4
        else:
            self.len_correction = 0

    def deserialize(self, response):
        if len(response) != len(self):
            raise ProtocolError(
                "message doesn't have expected length, expected %d bytes got %d"
                % (len(self), len(response))
            )
        if response[:4] != MESSAGE_HEADER:
            raise ProtocolError("Invalid message header")
        self.msg_id = response[4:8]
        self.seq_id = response[8:12]
        response = response[12:]
        if self.format_size == MESSAGE_SMALL or self.format_size == MESSAGE_LARGE:
            self.ack_id = response[:4]
            response = response[4:]
        if self.format_size != MESSAGE_SMALL:
            self.mac = response[:16]
            response = response[16:]

        response = self._parse_params(response)
        crc = response[:4]

        if response[4:] != MESSAGE_FOOTER:
            raise ProtocolError("Invalid message footer")

    def _parse_params(self, response):
        for p in self.params:
            my_val = response[: len(p)]
            p.deserialize(my_val)
            response = response[len(my_val) :]
        return response

    def __len__(self):
        arglen = sum(len(x) for x in self.params)
        return 34 + arglen + self.len_correction


class NodeAckSmallResponse(NodeResponse):
    """
    Acknowledge message without source MAC

    Response to: Any message
    """

    ID = b"0000"

    def __init__(self):
        super().__init__(MESSAGE_SMALL)


class NodeAckLargeResponse(NodeResponse):
    """
    Acknowledge message with source MAC

    Response to: Any message
    """

    ID = b"0000"

    def __init__(self):
        super().__init__(MESSAGE_LARGE)


class CirclePlusQueryResponse(NodeResponse):
    """
    TODO:

    Response to : ???
    """

    ID = b"0002"

    def __init__(self):
        super().__init__()
        self.channel = String(None, length=2)
        self.source_mac_id = String(None, length=16)
        self.extended_pan_id = String(None, length=16)
        self.unique_network_id = String(None, length=16)
        self.new_node_mac_id = String(None, length=16)
        self.pan_id = String(None, length=4)
        self.idx = Int(0, length=2)
        self.params += [
            self.channel,
            self.source_mac_id,
            self.extended_pan_id,
            self.unique_network_id,
            self.new_node_mac_id,
            self.pan_id,
            self.idx,
        ]

    def __len__(self):
        arglen = sum(len(x) for x in self.params)
        return 18 + arglen

    def deserialize(self, response):
        super().deserialize(self, response)
        # Clear first two characters of mac ID, as they contain part of the short PAN-ID
        self.new_node_mac_id.value = b"00" + self.new_node_mac_id.value[2:]


class CirclePlusQueryEndResponse(NodeResponse):
    """
    TODO:
        PWAckReplyV1_0
        <argument name="code" length="2"/>

    Response to : ???
    """

    ID = b"0003"

    def __init__(self):
        super().__init__()
        self.status = Int(0, 4)
        self.params += [self.status]

    def __len__(self):
        arglen = sum(len(x) for x in self.params)
        return 18 + arglen


class CirclePlusConnectResponse(NodeResponse):
    """
    CirclePlus connected to the network

    Response to : CirclePlusConnectRequest
    """

    ID = b"0005"

    def __init__(self):
        super().__init__()
        self.existing = Int(0, 2)
        self.allowed = Int(0, 2)
        self.params += [self.existing, self.allowed]

    def __len__(self):
        arglen = sum(len(x) for x in self.params)
        return 18 + arglen


class NodeJoinAvailableResponse(NodeResponse):
    """
    Message from an unjoined node to notify it is available to join a plugwise network

    Response to : <nothing>
    """

    ID = b"0006"


class StickInitResponse(NodeResponse):
    """
    Returns the configuration and status of the USB-Stick

    Optional:
    - circle_plus_mac
    - network_id

    <argument name="upYesNo" length="2"/>
    <argument name="extendedPanId" length="16" optional="1"/>
    <argument name="panId" length="4" optional="1"/>

    Response to: StickInitRequest
    """

    ID = b"0011"

    def __init__(self):
        super().__init__()
        self.unknown1 = Int(0, length=2)
        self.network_is_online = Int(0, length=2)
        self.circle_plus_mac = String(None, length=16)
        self.network_id = Int(0, length=4)
        self.unknown2 = Int(0, length=2)
        self.params += [
            self.unknown1,
            self.network_is_online,
            self.circle_plus_mac,
            self.network_id,
            self.unknown2,
        ]


class NodePingResponse(NodeResponse):
    """
    Ping response from node

    - incomingLastHopRssiTarget
    - lastHopRssiSource
    - timediffInMs

    Response to : NodePingRequest
    """

    ID = b"000E"

    def __init__(self):
        super().__init__()
        self.in_RSSI = Int(0, length=2)
        self.out_RSSI = Int(0, length=2)
        self.ping_ms = Int(0, length=4)
        self.params += [
            self.in_RSSI,
            self.out_RSSI,
            self.ping_ms,
        ]


class CirclePowerUsageResponse(NodeResponse):
    """
    Returns power usage as impulse counters for several different timeframes

    Response to : CirclePowerUsageRequest
    """

    ID = b"0013"

    def __init__(self):
        super().__init__()
        self.pulse_1s = Int(0, 4)
        self.pulse_8s = Int(0, 4)
        self.pulse_hour_consumed = Int(0, 8)
        self.pulse_hour_produced = Int(0, 8)
        self.nanosecond_offset = Int(0, 4)
        self.params += [
            self.pulse_1s,
            self.pulse_8s,
            self.pulse_hour_consumed,
            self.pulse_hour_produced,
            self.nanosecond_offset,
        ]


class CirclePlusScanResponse(NodeResponse):
    """
    Returns the MAC of a registered node at the specified memory address

    Response to: CirclePlusScanRequest
    """

    ID = b"0019"

    def __init__(self):
        super().__init__()
        self.node_mac = String(None, length=16)
        self.node_address = Int(0, length=2)
        self.params += [self.node_mac, self.node_address]


class NodeRemoveResponse(NodeResponse):
    """
    Returns conformation (or not) if node is removed from the Plugwise network
    by having it removed from the memory of the Circle+

    Response to: NodeRemoveRequest
    """

    ID = b"001D"

    def __init__(self):
        super().__init__()
        self.node_mac_id = String(None, length=16)
        self.status = Int(0, 2)
        self.params += [self.node_mac_id, self.status]


class NodeInfoResponse(NodeResponse):
    """
    Returns the status information of Node

    Response to: NodeInfoRequest
    """

    ID = b"0024"

    def __init__(self):
        super().__init__()
        self.datetime = DateTime()
        self.last_logaddr = LogAddr(0, length=8)
        self.relay_state = Int(0, length=2)
        self.hz = Int(0, length=2)
        self.hw_ver = String(None, length=12)
        self.fw_ver = UnixTimestamp(0)
        self.node_type = Int(0, length=2)
        self.params += [
            self.datetime,
            self.last_logaddr,
            self.relay_state,
            self.hz,
            self.hw_ver,
            self.fw_ver,
            self.node_type,
        ]


class CircleCalibrationResponse(NodeResponse):
    """
    returns the calibration settings of node

    Response to: CircleCalibrationRequest
    """

    ID = b"0027"

    def __init__(self):
        super().__init__()
        self.gain_a = Float(0, 8)
        self.gain_b = Float(0, 8)
        self.off_tot = Float(0, 8)
        self.off_noise = Float(0, 8)
        self.params += [self.gain_a, self.gain_b, self.off_tot, self.off_noise]


class CirclePlusRealTimeClockResponse(NodeResponse):
    """
    returns the real time clock of CirclePlus node

    Response to: CirclePlusRealTimeClockGetRequest
    """

    ID = b"003A"

    def __init__(self):
        super().__init__()

        self.time = RealClockTime()
        self.day_of_week = Int(0, length=2)
        self.date = RealClockDate()
        self.params += [self.time, self.day_of_week, self.date]


class CircleClockResponse(NodeResponse):
    """
    Returns the current internal clock of Node

    Response to: CircleClockGetRequest
    """

    ID = b"003F"

    def __init__(self):
        super().__init__()
        self.time = Time()
        self.day_of_week = Int(0, 2)
        self.unknown = Int(0, 2)
        self.unknown2 = Int(0, 4)
        self.params += [self.time, self.day_of_week, self.unknown, self.unknown2]


class CirclePowerBufferResponse(NodeResponse):
    """
    returns information about historical power usage
    each response contains 4 log buffers and each log buffer contains data for 1 hour

    Response to: CirclePowerBufferRequest
    """

    ID = b"0049"

    def __init__(self):
        super().__init__()
        self.logdate1 = DateTime()
        self.pulses1 = Int(0, 8)
        self.logdate2 = DateTime()
        self.pulses2 = Int(0, 8)
        self.logdate3 = DateTime()
        self.pulses3 = Int(0, 8)
        self.logdate4 = DateTime()
        self.pulses4 = Int(0, 8)
        self.logaddr = LogAddr(0, length=8)
        self.params += [
            self.logdate1,
            self.pulses1,
            self.logdate2,
            self.pulses2,
            self.logdate3,
            self.pulses3,
            self.logdate4,
            self.pulses4,
            self.logaddr,
        ]


class NodeAwakeResponse(NodeResponse):
    """
    A sleeping end device (SED: Scan, Sense, Switch) sends
    this message to announce that is awake. Awake types:
    - 0 : The SED joins the network for maintenance
    - 1 : The SED joins a network for the first time
    - 2 : The SED joins a network it has already joined, e.g. after reinserting a battery
    - 3 : When a SED switches a device group or when reporting values such as temperature/humidity
    - 4 : TODO: Unknown
    - 5 : A human pressed the button on a SED to wake it up

    Response to: <nothing>
    """

    ID = b"004F"

    def __init__(self):
        super().__init__()
        self.awake_type = Int(0, length=2)
        self.params += [self.awake_type]


class NodeSwitchGroupResponse(NodeResponse):
    """
    A sleeping end device (SED: Scan, Sense, Switch) sends
    this message to switch groups on/off when the configured
    switching conditions have been met.

    Response to: <nothing>
    """

    ID = b"0056"

    def __init__(self):
        super().__init__()
        self.group = Int(0, length=2)
        self.power_state = Int(0, length=2)
        self.params += [
            self.group,
            self.power_state,
        ]


class NodeFeaturesResponse(NodeResponse):
    """
    Returns supported features of node
    TODO: FeatureBitmask

    Response to: NodeFeaturesRequest
    """

    ID = b"0060"

    def __init__(self):
        super().__init__()
        self.features = Int(0, 16)
        self.params += [self.features]


class NodeJoinAckResponse(NodeResponse):
    """
    Notification mesage when node (re)joined existing network again.
    Sent when a SED (re)joins the network e.g. when you reinsert the battery of a Scan

    Response to: <nothing> or NodeAddRequest
    """

    ID = b"0061"

    def __init__(self):
        super().__init__()
        # sequence number is always FFFD


class NodeAckResponse(NodeResponse):
    """
    Acknowledge message in regular format
    Sent by nodes supporting plugwise 2.4 protocol version

    Response to:
    """

    ID = b"0100"

    def __init__(self):
        super().__init__()
        self.ack_id = Int(0, length=2)


class SenseReportResponse(NodeResponse):
    """
    Returns the current temperature and humidity of a Sense node.
    The interval this report is sent is configured by the 'SenseReportIntervalRequest' request

    Response to: <nothing>
    """

    ID = b"0105"

    def __init__(self):
        super().__init__()
        self.humidity = Int(0, length=4)
        self.temperature = Int(0, length=4)
        self.params += [self.humidity, self.temperature]


class CircleInitialRelaisStateResponse(NodeResponse):
    """
    Returns the initial relais state.

    Response to: CircleInitialRelaisStateRequest
    """

    ID = b"0139"

    def __init__(self):
        super().__init__()
        set_or_get = Int(0, length=2)
        relais = Int(0, length=2)
        self.params += [set_or_get, relais]



"""
Use of this source code is governed by the MIT license found in the LICENSE file.

Plugwise Circle node object
"""
import logging
from datetime import date, datetime, timedelta
from plugwise.constants import (
    ACK_OFF,
    ACK_ON,
    MAX_TIME_DRIFT,
    SENSOR_AVAILABLE,
    SENSOR_PING,
    SENSOR_POWER_USE,
    SENSOR_POWER_USE_LAST_8_SEC,
    SENSOR_POWER_CONSUMPTION_CURRENT_HOUR,
    SENSOR_POWER_CONSUMPTION_PREVIOUS_HOUR,
    SENSOR_POWER_CONSUMPTION_TODAY,
    SENSOR_POWER_CONSUMPTION_YESTERDAY,
    SENSOR_POWER_PRODUCTION_CURRENT_HOUR,
    SENSOR_POWER_PRODUCTION_PREVIOUS_HOUR,
    SENSOR_RSSI_IN,
    SENSOR_RSSI_OUT,
    SWITCH_RELAY,
    HA_SWITCH,
    HA_SENSOR,
    PULSES_PER_KW_SECOND,
)
from plugwise.node import PlugwiseNode

from plugwise.message import PlugwiseMessage
from plugwise.messages.requests import (
    CircleCalibrationRequest,
    CircleClockGetRequest,
    CircleClockSetRequest,
    CirclePowerBufferRequest,
    CirclePowerUsageRequest,
    CircleSwitchRelayRequest,
)
from plugwise.messages.responses import (
    CircleCalibrationResponse,
    CircleClockResponse,
    CirclePowerBufferResponse,
    CirclePowerUsageResponse,
    CirclePlusScanResponse,
    NodeAckLargeResponse,
)
from plugwise.util import Int


class PlugwiseCircle(PlugwiseNode):
    """provides interface to the Plugwise Circle nodes and base class for Circle+ nodes"""

    def __init__(self, mac, address, stick):
        super().__init__(mac, address, stick)
        self.categories = (HA_SWITCH, HA_SENSOR)
        self.sensors = (
            SENSOR_AVAILABLE["id"],
            SENSOR_PING["id"],
            SENSOR_POWER_USE["id"],
            SENSOR_POWER_USE_LAST_8_SEC["id"],
            SENSOR_POWER_CONSUMPTION_CURRENT_HOUR["id"],
            SENSOR_POWER_CONSUMPTION_PREVIOUS_HOUR["id"],
            SENSOR_POWER_CONSUMPTION_TODAY["id"],
            SENSOR_POWER_CONSUMPTION_YESTERDAY["id"],
            SENSOR_POWER_PRODUCTION_CURRENT_HOUR["id"],
            # SENSOR_POWER_PRODUCTION_PREVIOUS_HOUR["id"],
            SENSOR_RSSI_IN["id"],
            SENSOR_RSSI_OUT["id"],
        )
        self.switches = (SWITCH_RELAY["id"],)
        self.pulses_1s = None
        self.pulses_8s = None
        self.pulses_consumed_1h = None
        self.pulses_produced_1h = None
        self.calibration = False
        self._gain_a = None
        self._gain_b = None
        self._off_noise = None
        self._off_tot = None
        self.power_history = {}
        self.power_consumption_prev_hour = None
        self.power_consumption_today = None
        self.power_consumption_yesterday = None
        self._clock_offset = None
        self.get_clock(self.sync_clock)
        self._request_calibration()

    def _request_calibration(self, callback=None):
        """Request calibration info"""
        self.stick.send(
            CircleCalibrationRequest(self.mac),
            callback,
        )

    def _request_switch(self, state, callback=None):
        """Request to switch relay state and request state info"""
        self.stick.send(
            CircleSwitchRelayRequest(self.mac, state),
            callback,
        )

    def update_power_usage(self, callback=None):
        """Request power usage"""
        self.stick.send(
            CirclePowerUsageRequest(self.mac),
            callback,
        )

    def _on_message(self, message):
        """
        Process received message
        """
        if isinstance(message, CirclePowerUsageResponse):
            if self.calibration:
                self._response_power_usage(message)
                self.stick.logger.debug(
                    "Power update for %s, last update %s",
                    self.get_mac(),
                    str(self.last_update),
                )
            else:
                self.stick.logger.info(
                    "Received power update for %s before calibration information is known",
                    self.get_mac(),
                )
                self._request_calibration()
        elif isinstance(message, NodeAckLargeResponse):
            self._node_ack_response(message)
        elif isinstance(message, CircleCalibrationResponse):
            self._response_calibration(message)
        elif isinstance(message, CirclePowerBufferResponse):
            if self.calibration:
                self._response_power_buffer(message)
            else:
                self.stick.logger.debug(
                    "Received power buffer log for %s before calibration information is known",
                    self.get_mac(),
                )
                self._request_calibration()
        elif isinstance(message, CircleClockResponse):
            self._response_clock(message)
        else:
            self._circle_plus_message(message)

    def _circle_plus_message(self, message):
        pass

    def _process_scan_response(self, message):
        pass

    def get_relay_state(self) -> bool:
        """ Return last known relay state """
        return self._relay_state

    def set_relay_state(self, state: bool, callback=None):
        """ Switch relay """
        self._request_switch(state, callback)

    def get_power_usage(self):
        """
        Returns power usage during the last second in Watts
        Based on last received power usage information
        """
        if self.pulses_1s is not None:
            return self.pulses_to_kWs(self.pulses_1s) * 1000
        return None

    def get_power_usage_8_sec(self):
        """
        Returns power usage during the last 8 second in Watts
        Based on last received power usage information
        """
        if self.pulses_8s is not None:
            return self.pulses_to_kWs(self.pulses_8s, 8) * 1000
        return None

    def get_power_consumption_current_hour(self):
        """
        Returns the power usage during this running hour in kWh
        Based on last received power usage information
        """
        if self.pulses_consumed_1h is not None:
            return self.pulses_to_kWs(self.pulses_consumed_1h, 3600)
        return None

    def get_power_production_current_hour(self):
        """
        Returns the power production during this running hour in kWh
        Based on last received power usage information
        """
        if self.pulses_produced_1h is not None:
            return self.pulses_to_kWs(self.pulses_produced_1h, 3600)
        return None

    def get_power_consumption_prev_hour(self):
        """Returns power consumption during the previous hour in kWh"""
        return self.power_consumption_prev_hour

    def get_power_consumption_today(self):
        """Total power consumption during today in kWh"""
        return self.power_consumption_today

    def get_power_consumption_yesterday(self):
        """Total power consumption of yesterday in kWh"""
        return self.power_consumption_yesterday

    def _node_ack_response(self, message):
        """Process switch response message"""
        if message.ack_id == ACK_ON:
            if not self._relay_state:
                self.stick.logger.debug(
                    "Switch relay on for %s",
                    self.get_mac(),
                )
                self._relay_state = True
                self.do_callback(SWITCH_RELAY["id"])
        elif message.ack_id == ACK_OFF:
            if self._relay_state:
                self.stick.logger.debug(
                    "Switch relay off for %s",
                    self.get_mac(),
                )
                self._relay_state = False
                self.do_callback(SWITCH_RELAY["id"])
        else:
            self.stick.logger.debug(
                "Unmanaged _node_ack_response %s received for %s",
                str(message.ack_id),
                self.get_mac(),
            )

    def _response_power_usage(self, message):
        # Sometimes the circle returns -1 for some of the pulse counters
        # likely this means the circle measures very little power and is suffering from
        # rounding errors. Zero these out. However, negative pulse values are valid
        # for power producing appliances, like solar panels, so don't complain too loudly.

        # Power consumption last second
        if message.pulse_1s.value == -1:
            message.pulse_1s.value = 0
            self.stick.logger.debug(
                "1 sec power pulse counter for node %s has value of -1, corrected to 0",
                self.get_mac(),
            )
        self.pulses_1s = message.pulse_1s.value
        if message.pulse_1s.value != 0:
            if message.nanosecond_offset.value != 0:
                pulses_1s = (
                    message.pulse_1s.value
                    * (1000000000 + message.nanosecond_offset.value)
                ) / 1000000000
            else:
                pulses_1s = message.pulse_1s.value
            self.pulses_1s = pulses_1s
        else:
            self.pulses_1s = 0
        self.do_callback(SENSOR_POWER_USE["id"])
        # Power consumption last 8 seconds
        if message.pulse_8s.value == -1:
            message.pulse_8s.value = 0
            self.stick.logger.debug(
                "8 sec power pulse counter for node %s has value of -1, corrected to 0",
                self.get_mac(),
            )
        if message.pulse_8s.value != 0:
            if message.nanosecond_offset.value != 0:
                pulses_8s = (
                    message.pulse_8s.value
                    * (1000000000 + message.nanosecond_offset.value)
                ) / 1000000000
            else:
                pulses_8s = message.pulse_8s.value
            self.pulses_8s = pulses_8s
        else:
            self.pulses_8s = 0
        self.do_callback(SENSOR_POWER_USE_LAST_8_SEC["id"])
        # Power consumption current hour
        if message.pulse_hour_consumed.value == -1:
            message.pulse_hour_consumed.value = 0
            self.stick.logger.debug(
                "1 hour consumption power pulse counter for node %s has value of -1, corrected to 0",
                self.get_mac(),
            )
        self.pulses_consumed_1h = message.pulse_hour_consumed.value
        self.do_callback(SENSOR_POWER_CONSUMPTION_CURRENT_HOUR["id"])
        # Power produced current hour
        if message.pulse_hour_produced.value == -1:
            message.pulse_hour_produced.value = 0
            self.stick.logger.debug(
                "1 hour power production pulse counter for node %s has value of -1, corrected to 0",
                self.get_mac(),
            )
        self.pulses_produced_1h = message.pulse_hour_produced.value
        self.do_callback(SENSOR_POWER_PRODUCTION_CURRENT_HOUR["id"])

    def _response_calibration(self, message):
        """Store calibration properties"""
        for x in ("gain_a", "gain_b", "off_noise", "off_tot"):
            val = getattr(message, x).value
            setattr(self, "_" + x, val)
        self.calibration = True

    def pulses_to_kWs(self, pulses, seconds=1):
        """
        converts the amount of pulses to kWs using the calaboration offsets
        """
        if pulses == 0 or not self.calibration:
            return 0.0
        pulses_per_s = pulses / float(seconds)
        corrected_pulses = seconds * (
            (
                (((pulses_per_s + self._off_noise) ** 2) * self._gain_b)
                + ((pulses_per_s + self._off_noise) * self._gain_a)
            )
            + self._off_tot
        )
        calc_value = corrected_pulses / PULSES_PER_KW_SECOND / seconds
        # Fix minor miscalculations
        if calc_value < 0.001 and calc_value > -0.001:
            calc_value = 0.0
        return calc_value

    def _request_power_buffer(self, log_address=None, callback=None):
        """Request power log of specified address"""
        if log_address == None:
            log_address = self._last_log_address
        if log_address != None:
            if bool(self.power_history):
                # Only request last 2 power buffer logs
                self.stick.send(
                    CirclePowerBufferRequest(self.mac, log_address - 1),
                )
                self.stick.send(
                    CirclePowerBufferRequest(self.mac, log_address),
                    callback,
                )
            else:
                # Collect power history info of today and yesterday
                # Each request contains 4 hours except last request
                for req_log_address in range(log_address - 13, log_address):
                    self.stick.send(
                        CirclePowerBufferRequest(self.mac, req_log_address),
                    )
                self.stick.send(
                    CirclePowerBufferRequest(self.mac, log_address),
                    callback,
                )

    def _response_power_buffer(self, message):
        """returns information about historical power usage
        each response contains 4 log buffers and each log buffer contains data for 1 hour
        """
        if message.logaddr.value == self._last_log_address:
            self._last_log_collected = True
        # Collect logged power usage
        for i in range(1, 5):
            if getattr(message, "logdate%d" % (i,)).value != None:
                dt = getattr(message, "logdate%d" % (i,)).value
                if getattr(message, "pulses%d" % (i,)).value == 0:
                    self.power_history[dt] = 0.0
                else:
                    self.power_history[dt] = self.pulses_to_kWs(
                        getattr(message, "pulses%d" % (i,)).value, 3600
                    )
        # Cleanup history for more than 2 day's ago
        if len(self.power_history.keys()) > 48:
            for dt in list(self.power_history.keys()):
                if (dt + self.stick.timezone_delta - timedelta(hours=1)).date() < (
                    datetime.now().today().date() - timedelta(days=1)
                ):
                    del self.power_history[dt]
        # Recalculate power use counters
        last_hour_usage = 0
        today_power = 0
        yesterday_power = 0
        for dt in self.power_history:
            if (dt + self.stick.timezone_delta) == datetime.now().today().replace(
                minute=0, second=0, microsecond=0
            ):
                last_hour_usage = self.power_history[dt]
            if (
                dt + self.stick.timezone_delta - timedelta(hours=1)
            ).date() == datetime.now().today().date():
                today_power += self.power_history[dt]
            if (dt + self.stick.timezone_delta - timedelta(hours=1)).date() == (
                datetime.now().today().date() - timedelta(days=1)
            ):
                yesterday_power += self.power_history[dt]
        if self.power_consumption_prev_hour != last_hour_usage:
            self.power_consumption_prev_hour = last_hour_usage
            self.do_callback(SENSOR_POWER_CONSUMPTION_PREVIOUS_HOUR["id"])
        if self.power_consumption_today != today_power:
            self.power_consumption_today = today_power
            self.do_callback(SENSOR_POWER_CONSUMPTION_TODAY["id"])
        if self.power_consumption_yesterday != yesterday_power:
            self.power_consumption_yesterday = yesterday_power
            self.do_callback(SENSOR_POWER_CONSUMPTION_YESTERDAY["id"])

    def _response_clock(self, message):
        dt = datetime(
            datetime.now().year,
            datetime.now().month,
            datetime.now().day,
            message.time.value.hour,
            message.time.value.minute,
            message.time.value.second,
        )
        clock_offset = message.timestamp.replace(microsecond=0) - (
            dt + self.stick.timezone_delta
        )
        if clock_offset.days == -1:
            self._clock_offset = clock_offset.seconds - 86400
        else:
            self._clock_offset = clock_offset.seconds
        self.stick.logger.debug(
            "Clock of node %s has drifted %s sec",
            self.get_mac(),
            str(self._clock_offset),
        )

    def get_clock(self, callback=None):
        """ get current datetime of internal clock of Circle """
        self.stick.send(
            CircleClockGetRequest(self.mac),
            callback,
        )

    def set_clock(self, callback=None):
        """ set internal clock of CirclePlus """
        self.stick.send(
            CircleClockSetRequest(self.mac, datetime.utcnow()),
            callback,
        )

    def sync_clock(self, max_drift=0):
        """Resync clock of node if time has drifted more than MAX_TIME_DRIFT"""
        if self._clock_offset != None:
            if max_drift == 0:
                max_drift = MAX_TIME_DRIFT
            if (self._clock_offset > max_drift) or (self._clock_offset < -(max_drift)):
                self.stick.logger.info(
                    "Reset clock of node %s because time has drifted %s sec",
                    self.get_mac(),
                    str(self._clock_offset),
                )
                self.set_clock()



"""
Use of this source code is governed by the MIT license found in the LICENSE file.

Plugwise SED (Sleeping Endpoint Device) base object

TODO: 
- Expose awake state as sensor
- Set available state after 2 missed awake messages

"""

from plugwise.constants import (
    ACK_SLEEP_SET,
    NACK_SLEEP_SET,
    SED_AWAKE_BUTTON,
    SED_AWAKE_FIRST,
    SED_AWAKE_MAINTENANCE,
    SED_AWAKE_STARTUP,
    SED_AWAKE_STATE,
    SED_AWAKE_UNKNOWN,
    SED_CLOCK_INTERVAL,
    SED_CLOCK_SYNC,
    SED_MAINTENANCE_INTERVAL,
    SED_SLEEP_FOR,
    SED_STAY_ACTIVE,
    SENSOR_PING,
    SENSOR_RSSI_IN,
    SENSOR_RSSI_OUT,
)
from plugwise.node import PlugwiseNode
from plugwise.message import PlugwiseMessage
from plugwise.messages.responses import NodeAckLargeResponse, NodeAwakeResponse
from plugwise.messages.requests import (
    NodeInfoRequest,
    NodePingRequest,
    NodeSleepConfigRequest,
)


class NodeSED(PlugwiseNode):
    """provides base class for SED based nodes like Scan, Sense & Switch"""

    def __init__(self, mac, address, stick):
        super().__init__(mac, address, stick)
        self._SED_requests = {}
        self._maintenance_interval = SED_MAINTENANCE_INTERVAL
        self._new_maintenance_interval = None

    def is_sed(self) -> bool:
        """ Return if True if node SED (battery powered)"""
        return True

    def _on_message(self, message):
        """
        Process received message
        """
        if isinstance(message, NodeAwakeResponse):
            self._process_awake_response(message)
        elif isinstance(message, NodeAckLargeResponse):
            if message.ack_id == ACK_SLEEP_SET:
                self._maintenance_interval = self._new_maintenance_interval
            else:
                self._on_SED_message(message)
        else:
            self._on_SED_message(message)

    def _on_SED_message(self, message):
        pass

    def _process_awake_response(self, message):
        """" Process awake message"""
        self.stick.logger.debug(
            "Awake message type '%s' received from %s",
            str(message.awake_type.value),
            self.get_mac(),
        )
        if (
            message.awake_type.value == SED_AWAKE_MAINTENANCE
            or message.awake_type.value == SED_AWAKE_FIRST
            or message.awake_type.value == SED_AWAKE_STARTUP
            or message.awake_type.value == SED_AWAKE_BUTTON
        ):
            for request in self._SED_requests:
                (request_message, callback) = self._SED_requests[request]
                self.stick.logger.info(
                    "Send queued %s message to SED node %s",
                    request_message.__class__.__name__,
                    self.get_mac(),
                )
                self.stick.send(request_message, callback)
            self._SED_requests = {}
        else:
            if message.awake_type.value == SED_AWAKE_STATE:
                self.stick.logger.debug(
                    "Node %s awake for state change", self.get_mac()
                )
            else:
                self.stick.logger.info(
                    "Unknown awake message type (%s) received for node %s",
                    str(message.awake_type.value),
                    self.get_mac(),
                )

    def _queue_request(self, request_message, callback=None):
        """Queue request to be sent when SED is awake. Last message wins """
        self._SED_requests[request_message.ID] = (
            request_message,
            callback,
        )

    def _request_info(self, callback=None):
        """ Request info from node"""
        self._queue_request(
            NodeInfoRequest(self.mac),
            callback,
        )

    def ping(self, callback=None):
        """ Ping node"""
        if (
            self._callbacks.get(SENSOR_PING["id"])
            or self._callbacks.get(SENSOR_RSSI_IN["id"])
            or self._callbacks.get(SENSOR_RSSI_OUT["id"])
        ):
            self._queue_request(
                NodePingRequest(self.mac),
                callback,
            )
        else:
            self.stick.logger.debug(
                "Drop ping request for SED %s because no callback is registered",
                self.get_mac(),
            )

    def _wake_up_interval_accepted(self):
        """ Callback after wake up interval is received and accepted by SED """
        self._wake_up_interval = self._new_maintenance_interval

    def Configure_SED(
        self,
        stay_active=SED_STAY_ACTIVE,
        sleep_for=SED_SLEEP_FOR,
        maintenance_interval=SED_MAINTENANCE_INTERVAL,
        clock_sync=SED_CLOCK_SYNC,
        clock_interval=SED_CLOCK_INTERVAL,
    ):
        """Reconfigure the sleep/awake settings for a SED send at next awake of SED"""
        message = NodeSleepConfigRequest(
            self.mac,
            stay_active,
            maintenance_interval,
            sleep_for,
            clock_sync,
            clock_interval,
        )
        self._queue_request(message, self._wake_up_interval_accepted)
        self._new_maintenance_interval = maintenance_interval
        self.stick.logger.info(
            "Queue %s message to be send at next awake of SED node %s",
            message.__class__.__name__,
            self.get_mac(),
        )



""" Copyright (C) 2011 Sven Petai <hadara@bsd.ee>, use of this source code is governed by the MIT license found in the LICENSE file."""


class PlugwiseException(Exception):
    """Base error class for this Plugwise library"""


### Stick exceptions ###


class PortError(PlugwiseException):
    """Connection to USBstick failed"""


class StickInitError(PlugwiseException):
    """Initialization of USBstick failed"""


class NetworkDown(PlugwiseException):
    """Zigbee network not online"""


class CirclePlusError(PlugwiseException):
    """Connection to Circle+ node failed"""


class ProtocolError(PlugwiseException):
    """Error while decode received data"""


class TimeoutException(PlugwiseException):
    """Timeout expired while waiting for response from node"""


### Smile exceptions ###


class ConnectionFailedError(PlugwiseException):
    """Raised when unable to connect."""


class InvalidAuthentication(PlugwiseException):
    """Raised when unable to authenticate."""


class UnsupportedDeviceError(PlugwiseException):
    """Raised when device is not supported."""


class DeviceSetupError(PlugwiseException):
    """Raised when device is missing critical setup data."""


class DeviceTimeoutError(PlugwiseException):
    """Raised when device is not supported."""


class ErrorSendingCommandError(PlugwiseException):
    """Raised when device is not accepting the command."""


class ResponseError(PlugwiseException):
    """Raised when empty or error in response returned."""


class InvalidXMLError(PlugwiseException):
    """Raised when response holds incomplete or invalid XML data."""


class XMLDataMissingError(PlugwiseException):
    """Raised when xml data is empty."""



"""
Use of this source code is governed by the MIT license found in the LICENSE file.

Plugwise protocol helpers
"""
import binascii
import crcmod
import datetime
import logging
import re
import struct
import sys
from .exceptions import *
from .constants import (
    PLUGWISE_EPOCH,
    LOGADDR_OFFSET,
    UTF8_DECODE,
)


crc_fun = crcmod.mkCrcFun(0x11021, rev=False, initCrc=0x0000, xorOut=0x0000)


def validate_mac(mac):
    if not re.match("^[A-F0-9]+$", mac):
        return False
    try:
        _ = int(mac, 16)
    except ValueError:
        return False
    return True


def inc_seq_id(seq_id, value=1):
    """
    Increment sequence id by value

    :return: 4 bytes
    """
    temp_int = int(seq_id, 16) + value
    # Max seq_id = b'FFFC'
    # b'FFFD' reserved for 'NodeJoinAckResponse' message
    # b'FFFE' reserved for 'NodeSwitchGroupResponse' message
    # b'FFFF' reserved for 'NodeAwakeResponse' message
    if temp_int >= 65532:
        temp_int = 0
    temp_str = str(hex(temp_int)).lstrip("0x").upper()
    while len(temp_str) < 4:
        temp_str = "0" + temp_str
    return temp_str.encode()


def uint_to_int(val, octals):
    """compute the 2's compliment of int value val for negative values"""
    bits = octals << 2
    if (val & (1 << (bits - 1))) != 0:
        val = val - (1 << bits)
    return val


def int_to_uint(val, octals):
    """compute the 2's compliment of int value val for negative values"""
    bits = octals << 2
    if val < 0:
        val = val + (1 << bits)
    return val


def escape_illegal_xml_characters(xmldata):
    """Replace illegal &-characters."""
    return re.sub(r"&([^a-zA-Z#])", r"&amp;\1", xmldata)


def format_measure(measure):
    """Format measure to correct type."""
    try:
        measure = int(measure)
    except ValueError:
        try:
            if float(measure) < 10:
                measure = float(f"{round(float(measure), 2):.2f}")
            elif float(measure) >= 10 and float(measure) < 100:
                measure = float(f"{round(float(measure), 1):.1f}")
            elif float(measure) >= 100:
                measure = int(round(float(measure)))
        except ValueError:
            if measure == "on":
                measure = True
            elif measure == "off":
                measure = False
    return measure


def determine_selected(available, selected, schemas):
    """Determine selected schema from available schemas."""
    for schema_a, schema_b in schemas.items():
        available.append(schema_a)
        if schema_b:
            selected = schema_a
    return available, selected


def in_between(now, start, end):
    """Determine timing for schedules."""
    if start <= end:
        return start <= now < end
    return start <= now or now < end


class BaseType(object):
    def __init__(self, value, length):
        self.value = value
        self.length = length

    def serialize(self):
        return bytes(self.value, UTF8_DECODE)

    def deserialize(self, val):
        self.value = val

    def __len__(self):
        return self.length


class CompositeType(BaseType):
    def __init__(self):
        self.contents = []

    def serialize(self):
        return b"".join(a.serialize() for a in self.contents)

    def deserialize(self, val):
        for p in self.contents:
            myval = val[: len(p)]
            p.deserialize(myval)
            val = val[len(myval) :]
        return val

    def __len__(self):
        return sum(len(x) for x in self.contents)


class String(BaseType):
    pass


class Int(BaseType):
    def __init__(self, value, length=2):
        self.value = value
        self.length = length

    def serialize(self):
        fmt = "%%0%dX" % self.length
        return bytes(fmt % self.value, UTF8_DECODE)

    def deserialize(self, val):
        self.value = int(val, 16)
        mask = 1 << (self.length * 4 - 1)
        self.value = -(self.value & mask) + (self.value & ~mask)


class SInt(BaseType):
    def __init__(self, value, length=2):
        self.value = value
        self.length = length

    def negative(self, val, octals):
        """compute the 2's compliment of int value val for negative values"""
        bits = octals << 2
        if (val & (1 << (bits - 1))) != 0:
            val = val - (1 << bits)
        return val

    def serialize(self):
        fmt = "%%0%dX" % self.length
        return fmt % int_to_uint(self.value, self.length)

    def deserialize(self, val):
        self.value = self.negative(int(val, 16), self.length)


class UnixTimestamp(Int):
    def __init__(self, value, length=8):
        Int.__init__(self, value, length=length)

    def deserialize(self, val):
        Int.deserialize(self, val)
        self.value = datetime.datetime.fromtimestamp(self.value)


class Year2k(Int):
    """year value that is offset from the year 2000"""

    def deserialize(self, val):
        Int.deserialize(self, val)
        self.value += PLUGWISE_EPOCH


class DateTime(CompositeType):
    """datetime value as used in the general info response
    format is: YYMMmmmm
    where year is offset value from the epoch which is Y2K
    and last four bytes are offset from the beginning of the month in minutes
    """

    def __init__(self, year=0, month=0, minutes=0):
        CompositeType.__init__(self)
        self.year = Year2k(year - PLUGWISE_EPOCH, 2)
        self.month = Int(month, 2)
        self.minutes = Int(minutes, 4)
        self.contents += [self.year, self.month, self.minutes]

    def deserialize(self, val):
        CompositeType.deserialize(self, val)
        minutes = self.minutes.value
        hours = minutes // 60
        days = hours // 24
        hours -= days * 24
        minutes -= (days * 24 * 60) + (hours * 60)
        try:
            self.value = datetime.datetime(
                self.year.value, self.month.value, days + 1, hours, minutes
            )
        except ValueError:
            # debug(
            #    "encountered value error while attempting to construct datetime object"
            # )
            self.value = None


class Time(CompositeType):
    """time value as used in the clock info response"""

    def __init__(self, hour=0, minute=0, second=0):
        CompositeType.__init__(self)
        self.hour = Int(hour, 2)
        self.minute = Int(minute, 2)
        self.second = Int(second, 2)
        self.contents += [self.hour, self.minute, self.second]

    def deserialize(self, val):
        CompositeType.deserialize(self, val)
        self.value = datetime.time(
            self.hour.value, self.minute.value, self.second.value
        )


class IntDec(BaseType):
    def __init__(self, value, length=2):
        self.value = value
        self.length = length

    def serialize(self):
        fmt = "%%0%dd" % self.length
        return bytes(fmt % self.value, UTF8_DECODE)

    def deserialize(self, val):
        self.value = val.decode(UTF8_DECODE)


class RealClockTime(CompositeType):
    """time value as used in the realtime clock info response"""

    def __init__(self, hour=0, minute=0, second=0):
        CompositeType.__init__(self)
        self.hour = IntDec(hour, 2)
        self.minute = IntDec(minute, 2)
        self.second = IntDec(second, 2)
        self.contents += [self.second, self.minute, self.hour]

    def deserialize(self, val):
        CompositeType.deserialize(self, val)
        self.value = datetime.time(
            int(self.hour.value), int(self.minute.value), int(self.second.value),
        )


class RealClockDate(CompositeType):
    """date value as used in the realtime clock info response"""

    def __init__(self, day=0, month=0, year=0):
        CompositeType.__init__(self)
        self.day = IntDec(day, 2)
        self.month = IntDec(month, 2)
        self.year = IntDec(year - PLUGWISE_EPOCH, 2)
        self.contents += [self.day, self.month, self.year]

    def deserialize(self, val):
        CompositeType.deserialize(self, val)
        self.value = datetime.date(
            int(self.year.value) + PLUGWISE_EPOCH,
            int(self.month.value),
            int(self.day.value),
        )


class Float(BaseType):
    def __init__(self, value, length=4):
        self.value = value
        self.length = length

    def deserialize(self, val):
        hexval = binascii.unhexlify(val)
        self.value = struct.unpack("!f", hexval)[0]


class LogAddr(Int):
    def serialize(self):
        return bytes("%08X" % ((self.value * 32) + LOGADDR_OFFSET), UTF8_DECODE)

    def deserialize(self, val):
        Int.deserialize(self, val)
        self.value = (self.value - LOGADDR_OFFSET) // 32
